import argparse
from pathlib import Path
from typing import Any, Dict, List, Tuple

from . import utils, utils_cwl, utils_graphs
from .wic_types import (GraphReps, InternalOutputs, Namespaces, StepId, Tool, Tools,
                        WorkflowInputs, Yaml)

# NOTE: This must be initialized in main.py and/or cwl_subinterpreter.py
renaming_conventions: List[Tuple[str, str]] = []


def types_match(in_type: Any, out_type: Any) -> bool:
    if in_type == out_type:
        return True
    if isinstance(in_type, list) and not isinstance(out_type, list):
        return any([x == out_type for x in in_type])
    if isinstance(out_type, list) and not isinstance(in_type, list):
        # If only some of the output types match the input type, and
        # if at runtime the actual output is one of those matches,
        # then great! But we can't know that at compile time.
        # If we want to guarantee this won't fail (due to format mismatch)
        # at runtime then use all, not any. However, using all excludes many
        # common use cases (format: ["null", ...]) and there are plenty of
        # other ways to fail at runtime, so for now let's allow it.
        # (If inference makes the 'wrong' choice, just use an explicit edge!)
        return any([x == in_type for x in out_type])
    if isinstance(in_type, list) and isinstance(out_type, list):
        # Same comment here
        return any([x in in_type for x in out_type])
    return False


def perform_edge_inference(args: argparse.Namespace,
                           tools: Tools,
                           tools_lst: List[Tool],
                           steps_keys: List[str],
                           yaml_stem: str,
                           i: int,
                           steps: List[Yaml],
                           arg_key: str,
                           graph: GraphReps,
                           is_root: bool,
                           namespaces: Namespaces,
                           vars_workflow_output_internal: InternalOutputs,
                           input_mapping: Dict[str, List[str]],
                           output_mapping: Dict[str, str],
                           inputs_workflow: WorkflowInputs,
                           in_name: str,
                           in_name_in_inputs_file_workflow: bool,
                           arg_key_in_yaml_tree_inputs: bool,
                           insertions: List[StepId],
                           wic_steps: Yaml,
                           testing: bool) -> Yaml:
    """This function implements the core edge inference feature.
    NOTE: steps[i], vars_workflow_output_internal, inputs_workflow are mutably updated.

    Args:
        args (argparse.Namespace): The command line arguments
        tools (Tools): The CWL CommandLineTool definitions found using get_tools_cwl()
        tools_lst (List[Tool]): A list of the CWL CommandLineTools or compiled subworkflows for the current workflow.
        steps_keys (List[str]): The name of each step in the current CWL workflow
        yaml_stem (str): The name (filename without extension) of the current CWL workflow
        i (int): The (zero-based) step number w.r.t. the current subworkflow.\n
        Since we are trying to infer inputs from previous outputs, this will not\n
        perform any inference (again, w.r.t. the current subworkflow) if i == 0.
        steps (List[Yaml]): The steps: tag of the current CWL workflow
        arg_key (str): The name of the CWL input tag that needs a concrete input value inferred
        graph (GraphReps): A tuple of a GraphViz DiGraph and a networkx DiGraph
        is_root (bool): True if this is the root workflow (for debugging only)
        namespaces (Namespaces): Specifies the path in the AST of the current subworkflow
        vars_workflow_output_internal (InternalOutputs): Keeps track of output\n
        variables which are internal to the root workflow, but not necessarily to subworkflows.
        input_mapping (Dict[str, List[str]]): Maps workflow inputs to workflow step inputs, recursively namespaced.
        output_mapping (Dict[str, str]): Maps workflow outputs to workflow step outputs, recursively namespaced.
        inputs_workflow (WorkflowInputs): Keeps track of CWL inputs: variables for the current workflow.
        in_name (str): The input name
        in_name_in_inputs_file_workflow (bool): Used to determine whether\n
        failure to find a match should be considered an error.
        arg_key_in_yaml_tree_inputs (bool): Determines whether at least one level of recursion has been performed.
        insertions (List[StepId]): If exact inference fails, a list of possible steps to automatically insert is stored here.
        wic_steps (Yaml): The metadata associated with the given workflow.
        testing: Used to disable some optional features which are unnecessary for testing.

    Returns:
        Yaml: steps[i] with the input tag arg_key updated with an inferred input value.
    """
    # Use in_name_in_inputs_file_workflow: bool so that at the call site, we don't
    # have to question whether or not this function modifies inputs_file_workflow
    # gmx_trjconv_str assumes gro, but other tools assume tpr; For now,
    # instead of figuring out how to replace('structure_', 'gro_') for
    # gmx_trjconv_str only, just require the user to specify filename.
    step_key = steps_keys[i]
    tool_i = tools_lst[i]
    step_name_i = utils.step_name_str(yaml_stem, i, step_key)

    in_tool = tool_i.cwl['inputs']
    in_dict = utils_cwl.copy_cwl_input_output_dict(in_tool[arg_key], True)

    if arg_key in steps[i].get('scatter', []):
        # Promote scattered input types to arrays
        in_dict['type'] = {'type': 'array', 'items': in_dict['type']}

    in_formats = []
    if 'format' in in_tool[arg_key]:
        in_formats = in_tool[arg_key]['format']
        in_dict['format'] = in_formats
    # print('step_name_i, arg_key, in_formats', step_name_i, arg_key, in_formats)
    format_matches_all = []
    attempted_matches_all = []
    break_inference = False
    for j in range(0, i)[::-1]:  # Reverse order!
        wic_step_j = wic_steps.get(f'({j+1}, {steps_keys[j]})', {})
        tool_j = tools_lst[j]
        out_tool = tool_j.cwl['outputs']
        # NOTE: The outputs of a CommandLineTool are all made available
        # simultaneously. Although that is technically also true for subworkflows,
        # the CommandLineTools within the subworkflow are certainly ordered,
        # and so we definitely want to use reverse order here. As mentioned,
        # it doesn't necessarily make sense for CommandLineTools, but the
        # important thing is that we just define the order for users some way.
        out_keys = list(tool_j.cwl['outputs'])[::-1]  # Reverse order!
        format_matches = []
        attempted_matches = []
        inference_rules = get_inference_rules(wic_step_j, Path(steps_keys[j]).stem)
        namespace_emb_last_break = ''
        for out_key in out_keys:
            namespaces_embedded = out_key.split('___')
            namespace_emb_last = '' if len(namespaces_embedded) <= 1 else namespaces_embedded[:-1][-1]  # -2?
            if break_inference and namespace_emb_last != namespace_emb_last_break:
                break  # Only break once the namespace changes, i.e. on the next step
            inference_rule = inference_rules.get(out_key, 'default')
            # Apply 'continue' rule before iteration, to prevent matching
            # TODO: This currently causes an infinite loop.
            # if inference_rule == 'continue':
            #    continue

            out_dict = utils_cwl.copy_cwl_input_output_dict(out_tool[out_key])

            if 'scatter' in steps[j]:
                # Promote scattered output types to arrays
                out_dict['type'] = {'type': 'array', 'items': out_dict['type']}

            out_format = ''
            if 'format' in out_tool[out_key]:
                out_format = out_tool[out_key]['format']
                out_dict['format'] = out_format
            # if out_format == '':
            #    #print('Warning! No output format! Cannot possibly match!')
            #    print('Warning! No output format! Will match anything!')
            #    print(f'out_key {out_key}')
            # print('out_key, out_format, rule', out_key, out_format, inference_rule)
            attempted_matches.append((out_key, out_format))
            # Great! We found an 'exact' type and format match.
            if types_match(in_dict['type'], out_dict['type']):  # First we have to match the types.
                if in_formats:
                    # Then, if we have an input format or formats, the output format has to match.
                    if out_format == in_formats or out_format in in_formats:
                        format_matches.append((out_key, out_format))
                else:
                    # Otherwise, formats are optional and we match only on type.
                    format_matches.append((out_key, out_format))

            # Apply 'break' rule after iteration, to allow matching
            if inference_rule == 'break':
                break_inference = True
                namespace_emb_last_break = namespace_emb_last

        format_matches_all.append(format_matches)
        attempted_matches_all.append(attempted_matches)
        # Most log files just have format_2330 "Textual format", but this can
        # conflict with other structured text files that also have format_2330.
        # Unfortunately, the edam formats are not very finely curated, so they
        # are not very 'exact'. For now, we can perform additional matching
        # based on naming conventions and/or we can simply exclude log files
        # (which are not usually parsed or otherwise used as inputs).
        # Eventually, we will want to improve the format curation.
        # NOTE: Use underscores to prevent excluding e.g. 'topology'
        # This isn't great, but works for now (until someone uses '_log_' ...)
        format_matches = [x for x in format_matches if not '_log_' in x[0]]
        # print('format_matches', format_matches)
        if not len(format_matches) == 0:
            # By default, simply choose the first (i.e. most-recent) matching format
            out_key = format_matches[0][0]

            if args.inference_use_naming_conventions:  # default False
                if len(format_matches) == 1:
                    # Great! We found a unique format match.
                    out_key = format_matches[0][0]
                else:
                    name_matches = []
                    # NOTE: The biobb CWL files do not use consistent naming
                    # conventions, so we need to perform some renamings here.
                    # Eventually, the CWL files themselves should be fixed.
                    arg_key_no_namespace = arg_key.split('___')[-1]
                    arg_key_renamed = arg_key_no_namespace.replace('input_', '')
                    for name1, name2 in renaming_conventions:
                        arg_key_renamed = arg_key_renamed.replace(name1, name2)

                    for out_key, out_format in format_matches:
                        out_key_no_namespace = out_key.split('___')[-1]
                        out_key_renamed = out_key_no_namespace.replace('output_', '')
                        if arg_key_renamed == out_key_renamed:
                            name_matches.append((out_key, out_format))

                    if len(name_matches) == 0:
                        # s = f"""Found multiple outputs with compatible types and formats
                        # (but no matching names) for input {arg_key}"""
                        # print(s)
                        # for m in format_matches:
                        #    print(m)
                        # print(f'Arbitrarily choosing the first match {format_matches[0][0]}')
                        out_key = format_matches[0][0]
                    elif len(name_matches) == 1:
                        # NOTE: This clause currently causes problems with insertions.
                        # Specifically, we want to convert from format A to B, perform the calculation
                        # in format B, then convert the results back to format A. However, if the
                        # naming conventions of the result do not match the second insertion
                        # (but DO match the first insertion), the files will be directly converted
                        # from A to B to A, thus skipping the calculation in B entirely!
                        # Great! We found a unique match.
                        out_key = name_matches[0][0]
                        # print('unique match', out_key)
                    else:
                        # s = f"""Found multiple outputs with compatible types and formats
                        # (and multiple matching names) for input {arg_key}"""
                        # print(s)
                        # for m in name_matches:
                        #    print(m)
                        # print(f'Arbitrarily choosing the first match {name_matches[0][0]}')
                        out_key = name_matches[0][0]

            # print('match!', j)  # We found a match!
            # Generate a new namespace for out_key using the step number and add to inputs
            step_name_j = utils.step_name_str(yaml_stem, j, steps_keys[j])

            # We also need to keep track of the 'internal' output variables
            if tool_j.cwl['class'] == 'Workflow':
                vars_workflow_output_internal.append(out_key)
            else:
                vars_workflow_output_internal.append(f'{step_name_j}/{out_key}')

            arg_val = f'{step_name_j}/{out_key}'
            arg_keyval = {arg_key: arg_val}
            steps_i = utils_cwl.add_yamldict_keyval_in(steps[i], step_key, arg_keyval)
            # print(f'inference i {i} y arg_key {arg_key}')

            arg_keys = [in_name] if in_name in input_mapping else [arg_key]
            arg_keys = utils.get_input_mappings(input_mapping, arg_keys, arg_key_in_yaml_tree_inputs)

            out_key = utils.get_output_mapping(output_mapping, out_key)

            nss_embedded1 = out_key.split('___')[:-1]

            # NOTE: This if statement is unmotivated and probably masking some other bug, but it works.
            if out_key.startswith('___'.join(namespaces + [step_name_j])):
                nss1 = nss_embedded1
            elif out_key.startswith(step_name_j):
                nss1 = namespaces + nss_embedded1
            else:
                nss1 = namespaces + [step_name_j] + nss_embedded1

            for arg_key_ in arg_keys:
                # Determine which head and tail node to use for the new edge
                # First we need to extract the embedded namespaces
                nss_embedded2 = arg_key_.split('___')[:-1]

                # NOTE: This if statement is unmotivated and probably masking some other bug, but it works.
                if arg_key_.startswith('___'.join(namespaces + [step_name_i])):
                    nss2 = nss_embedded2
                elif arg_key_.startswith(step_name_i):
                    nss2 = namespaces + nss_embedded2
                else:
                    nss2 = namespaces + [step_name_i] + nss_embedded2

                # TODO: check this
                out_key_no_namespace = out_key.split('___')[-1]
                label = out_key_no_namespace if tool_j.cwl['class'] == 'Workflow' else out_key
                utils_graphs.add_graph_edge(args, graph, nss1, nss2, label)

            return steps_i  # Short circuit

        # Stop performing inference if the inference rule is 'break'
        if break_inference:
            break

    # If no match yet, we can look for a potential steps to insert automatically.
    # NOTE: PLEASE READ docs/advanced.md#program-synthesis
    out_formats = [out_format for attempted_matches in attempted_matches_all
                   for (out_key_, out_format) in attempted_matches]
    # print('out_formats', out_formats)
    for in_format in in_formats:
        for out_format in out_formats:
            # Obviously we don't need an insertion if the file formats are the same.
            if in_format == out_format:
                continue

            for step_id, tool in tools.items():
                # For now, let's restrict to a whitelist.
                # Otherwise, there are way too many solutions.
                # (In principle, this can be used to insert arbitrary subworkflows.)
                if not step_id.stem.startswith('insert_steps_automatically_'):
                    continue

                in_tool = tool.cwl['inputs']
                tool_in_formats = [arg_val['format'] for arg_key, arg_val in in_tool.items() if 'format' in arg_val]
                tool_in_formats_flat = utils.flatten(tool_in_formats)

                out_tool = tool.cwl['outputs']
                tool_out_formats = [out_val['format'] for out_key, out_val in out_tool.items() if 'format' in out_val]

                # NOTE: Ideally, the second condition should really check that
                # ALL required inputs for the intermediate tool match with the
                # an element of out_formats. Since we cannot easily do that here,
                # we need to tentatively insert tool into the AST and re-compile.
                # However, that can easily fail (i.e. if there is one transitive
                # match). See docs/algorithms.md for more details.
                if in_format in tool_out_formats and out_format in tool_in_formats_flat:
                    # We may have found an insertion.
                    # print('step_id.stem, in_format, out_format:', step_id.stem, in_format, out_format)
                    insertions.append(step_id)

    match = False
    if not match:
        # print(f'inference i {i} n arg_key {arg_key}')
        # Use triple underscore for namespacing so we can split later
        in_name = f'{step_name_i}___{arg_key}'  # {step_name_i}_input___{arg_key}

        # This just means we need to defer to the parent workflow.
        # There will actually be an error only if no parent supplies an
        # input value and thus there is no entry in inputs_file_workflow.
        if is_root and not in_name_in_inputs_file_workflow and not testing:
            print('Error! No match found for input', i + 1, step_key, arg_key)
            # NOTE: The following print statement is a bit misleading because
            # we don't print out any attempted matches in the recursive case.
            print('number of attempted matches ', len(utils.flatten(attempted_matches_all)))
            for a_m in attempted_matches_all:
                for m in a_m:
                    print(m)

        # Add an input name to this subworkflow (only). Do not add to
        # inputs_file_workflow because this may be an internal input,
        # i.e. it may simply be split across subworkflow boundaries and
        # the input may be supplied in one of the parent workflows.
        # We also do not (necessarily) need to explicitly pass this list
        # to the parent workflow since we are compiling the 'in' tag here,
        # which should match in the parent workflow.
        inputs_workflow.update({in_name: in_dict})

        arg_keyval = {arg_key: in_name}
        steps_i = utils_cwl.add_yamldict_keyval_in(steps[i], step_key, arg_keyval)
        return steps_i

    # Add an explicit return statement here so mypy doesn't complain.
    # (mypy's static analysis cannot determine that this is dead code.)
    return steps[i]


def get_inference_rules(wic: Yaml, step_key_parent: str) -> Dict[str, str]:
    """Recursively traverses the wic: metadata annotation AST and extracts any inference rules.\n
    See docs/userguide.md for more information.

    Args:
        wic (Yaml): The contents of the wic: metadata annotations tag (if any)
        step_key_parent (str): The name of one of the steps in the current workflow.

    Returns:
        Dict[str, str]: A dictionary of the inference rules for the workflow step named step_key_parent.
    """
    # NOTE: Here, we simply return all inference rules. The call site then
    # determines whether to apply any of the rules by doing a lookup.
    if 'steps' in wic.get('wic', {}):
        wic_steps = wic['wic']['steps']
        rules = {}
        for keystr, wic_child in wic_steps.items():
            (step_num, step_key) = utils.parse_int_string_tuple(keystr)
            namespace = utils.step_name_str(Path(step_key_parent).stem, step_num - 1, step_key)
            rules_child = get_inference_rules(wic_child, step_key)
            for rule_key, rule_val in rules_child.items():
                rules[namespace + '___' + rule_key] = rule_val  # Namespaceing
        return rules
    if 'inference' in wic.get('wic', {}):
        rules = wic['wic']['inference']
        return rules
    return {}



import argparse
import copy
from pathlib import Path
from typing import Any, Dict, List
import yaml

from . import utils
from .wic_types import (GraphReps, InternalOutputs, Namespaces, Tool, Tools,
                        WorkflowOutputs, Yaml, StepId)


def maybe_add_requirements(yaml_tree: Yaml, steps_keys: List[str],
                           wic_steps: Yaml, subkeys: List[str]) -> None:
    """Adds any necessary CWL requirements

    Args:
        yaml_tree (Yaml): A tuple of name and yml AST
        steps_keys (List[str]): The name of each step in the current CWL workflow
        wic_steps (Yaml): The metadata associated with the workflow steps
        subkeys (List[str]): The keys associated with subworkflows
    """
    subwork = []
    scatter = []
    stepinp = []
    jsreq = []
    for i, step_key in enumerate(steps_keys):
        sub_wic = wic_steps.get(f'({i+1}, {step_key})', {})

        if 'scatter' in yaml_tree['steps'][i]:
            scatter = ['ScatterFeatureRequirement']
        if 'when' in yaml_tree['steps'][i]:
            jsreq = ['InlineJavascriptRequirement']

        in_step = yaml_tree['steps'][i].get('in')
        sub_wic_copy = copy.deepcopy(sub_wic)
        if 'wic' in sub_wic_copy:
            del sub_wic_copy['wic']
        if (utils.recursively_contains_dict_key('valueFrom', in_step) or
                utils.recursively_contains_dict_key('valueFrom', sub_wic_copy)):
            stepinp = ['StepInputExpressionRequirement', 'InlineJavascriptRequirement']

    if not subkeys == []:
        subwork = ['SubworkflowFeatureRequirement']

    reqs = subwork + scatter + stepinp + jsreq
    if reqs:
        reqsdict: Dict[str, Dict] = {r: {} for r in set(reqs)}
        if 'requirements' in yaml_tree:
            new_reqs = dict(list(yaml_tree['requirements'].items()) + list(reqsdict.items()))
            yaml_tree['requirements'].update(new_reqs)
        else:
            yaml_tree['requirements'] = reqsdict


def add_yamldict_keyval_in(steps_i: Yaml, step_key: str, keyval: Yaml) -> Yaml:
    """Convenience function used to (mutably) merge two Yaml dicts.

    Args:
        steps_i (Yaml): A partially-completed Yaml dict representing a step in a CWL workflow
        step_key (str): The name of the step in a CWL workflow
        keyval (Yaml): A Yaml dict with additional details to be merged into the first Yaml dict

    Returns:
        Yaml: The first Yaml dict with the second Yaml dict merged into it.
    """
    # TODO: Check whether we can just use deepmerge.merge()
    if steps_i:
        if 'in' in steps_i:
            new_keys = dict(list(steps_i['in'].items()) + list(keyval.items()))
            new_keyvals = dict([(k, v) if k != 'in' else (k, new_keys) for k, v in steps_i.items()])
        else:
            new_keys = keyval
            new_keyvals = dict(list(steps_i.items()) + [('in', new_keys)])
        steps_i.update(new_keyvals)
    else:
        steps_i = {step_key: {'in': keyval}}
    return steps_i


def add_yamldict_keyval_out(steps_i: Yaml, step_key: str, strs: List[str]) -> Yaml:
    """Convenience function used to (mutably) merge two Yaml dicts.

    Args:
        steps_i (Yaml): A partially-completed Yaml dict representing a step in a CWL workflow
        step_key (str): The name of the step in a CWL workflow
        keyval (Yaml): A Yaml dict with additional details to be merged into the first Yaml dict

    Returns:
        Yaml: The first Yaml dict with the second Yaml dict merged into it.
    """
    # TODO: Check whether we can just use deepmerge.merge()
    if steps_i:
        if 'out' in steps_i:
            new_strs = steps_i['out'] + strs
            new_keyvals = dict([(k, v) if k != 'out' else (k, new_strs) for k, v in steps_i.items()])
        else:
            new_keyvals = dict(list(steps_i.items()) + [('out', strs)])
        steps_i.update(new_keyvals)
    else:
        steps_i = {step_key: {'out': strs}}
    return steps_i


def get_workflow_outputs(args: argparse.Namespace,
                         namespaces: Namespaces,
                         is_root: bool,
                         yaml_stem: str,
                         steps: List[Yaml],
                         outputs_workflow: WorkflowOutputs,
                         vars_workflow_output_internal: InternalOutputs,
                         graph: GraphReps,
                         tools_lst: List[Tool],
                         step_node_name: str,
                         tools: Tools) -> Dict[str, Dict[str, str]]:
    """Chooses a subset of the CWL outputs: to actually output

    Args:
        args (argparse.Namespace): The command line arguments
        namespaces (Namespaces): Specifies the path in the AST of the current subworkflow
        is_root (bool): True if this is the root workflow
        yaml_stem (str): The name of the current subworkflow (stem of the yaml filepath)
        steps (List[Yaml]): The steps: tag of a CWL workflow
        outputs_workflow (WorkflowOutputs): Contains the contents of the out: tags for each step.
        vars_workflow_output_internal (InternalOutputs): Keeps track of output\n
        variables which are internal to the root workflow, but not necessarily to subworkflows.
        graph (GraphReps): A tuple of a GraphViz DiGraph and a networkx DiGraph
        tools_lst (List[Tool]): A list of the CWL CommandLineTools or compiled subworkflows for the current workflow.
        step_node_name (str): The namespaced name of the current step
        tools (Tools): The CWL CommandLineTool definitions found using get_tools_cwl()

    Returns:
        Dict[str, Dict[str, str]]: The actual outputs to be specified in the generated CWL file
    """
    # Add the outputs of each step to the workflow outputs
    workflow_outputs = {}
    steps_keys = utils.get_steps_keys(steps)
    for i, step_key in enumerate(steps_keys):
        tool_i = tools_lst[i].cwl
        step_name_i = utils.step_name_str(yaml_stem, i, step_key)
        step_id = StepId(Path(step_key).stem, 'global')
        step_name_or_key = step_name_i if step_key.endswith('.wic') \
            or step_id in tools \
            or Path(step_key).stem == Path(tools_lst[i].run_path).stem else step_key
        # step_name_or_key = step_name_i if stepid in tools else step_key
        out_keys = steps[i]['out']
        for out_key in out_keys:
            out_var = f'{step_name_or_key}/{out_key}'
            # Avoid duplicating intermediate outputs in GraphViz
            out_key_no_namespace = out_key.split('___')[-1]
            if args.graph_show_outputs:
                vars_nss = [var.replace('/', '___') for var in vars_workflow_output_internal]
                case1 = (tool_i['class'] == 'Workflow') and (not out_key in vars_nss)
                # Avoid duplicating outputs from subgraphs in parent graphs.
                namespaced_output_name = '___'.join(namespaces + [step_name_or_key, out_key])
                lengths_off_by_one = (len(step_node_name.split('___')) + 1 == len(namespaced_output_name.split('___')))
                # TODO: check is_root here
                case1 = case1 and not is_root and lengths_off_by_one
                case2 = (tool_i['class'] == 'CommandLineTool') and (not out_var in vars_workflow_output_internal)
                if case1 or case2:
                    graph_gv = graph.graphviz
                    graph_nx = graph.networkx
                    graphdata = graph.graphdata
                    attrs = {'label': out_key_no_namespace, 'shape': 'box',
                             'style': 'rounded, filled', 'fillcolor': 'lightyellow'}
                    graph_gv.node(namespaced_output_name, **attrs)
                    font_edge_color = 'black' if args.graph_dark_theme else 'white'
                    if args.graph_label_edges:
                        graph_gv.edge(step_node_name, namespaced_output_name, color=font_edge_color,
                                      label=out_key_no_namespace)  # Is labeling necessary?
                    else:
                        graph_gv.edge(step_node_name, namespaced_output_name, color=font_edge_color)
                    graph_nx.add_node(namespaced_output_name)
                    graph_nx.add_edge(step_node_name, namespaced_output_name)
                    graphdata.nodes.append((namespaced_output_name, attrs))
                    graphdata.edges.append((step_node_name, namespaced_output_name, {}))
            # NOTE: Unless we are in the root workflow, we always need to
            # output everything. This is because while we are within a
            # subworkflow, we do not yet know if a subworkflow output will be used as
            # an input in a parent workflow (either explicitly, or using inference).
            # Thus, we have to output everything.
            # However, once we reach the root workflow, we can choose whether
            # we want to pollute our output directory with intermediate files.
            # (You can always enable --provenance to get intermediate files.)
            # NOTE: Remove is_root for now because in test_cwl_embedding_independence,
            # we recompile all subworkflows as if they were root.
            # Exclude intermediate 'output' files.
            if out_var in vars_workflow_output_internal:  # and is_root
                continue
            out_name = f'{step_name_or_key}___{out_key}'  # Use triple underscore for namespacing so we can split later
            # print('out_name', out_name)

        for out_key, out_dict in outputs_workflow[i].items():
            out_dict['type'] = canonicalize_type(out_dict['type'])
            if 'scatter' in steps[i]:
                # Promote scattered output types to arrays
                out_dict['type'] = {'type': 'array', 'items': out_dict['type']}

            out_name = f'{step_name_or_key}___{out_key}'  # Use triple underscore for namespacing so we can split later
            out_var = f'{step_name_or_key}/{out_key}'
            workflow_outputs.update({out_name: {**out_dict, 'outputSource': out_var}})
        # print('workflow_outputs', workflow_outputs)
    # NOTE: The fix_conflicts 'feature' of cwltool prevents files from being
    # overwritten by appending _2, _3 etc.
    # The problem is that these renamed files now no longer match the glob
    # patterns in the outputBinding tags, thus they are not copied to the
    # final output folder in relocateOutputs() and/or stage_files().
    # Note that this error is not detected using --validate.
    # One workaround is to simply output all files.
    # TODO: glob "." is still returning null; need to use InitialWorkDirRequirement??
    output_all = {'output_all':
                  {'type':
                   {'type': 'array',
                    'items': ['Directory', 'File']},
                   'outputBinding': {'glob': '\".\"'},
                   'format': 'edam:format_2330'}}  # 'Textual format'
    # This crashes toil-cwl-runner, but not cwltool.
    # workflow_outputs.update(output_all) # type: ignore
    return workflow_outputs


def canonicalize_type(type_obj: Any) -> Any:
    """Recursively desugars the CWL type: field into a canonical normal form.\n
    In particular, CWL automatically desugars File[] into {'type': 'array', 'items': File},
    but File[][] causes a syntax error! Etc.

    Args:
        type_obj (Any): An object that is a syntactic hodgepodge of valid CWL types.

    Returns:
        Any: The JSON canonical normal form associated with type_obj
    """
    if isinstance(type_obj, str):
        if len(type_obj) >= 1 and type_obj[-1:] == '?':
            return ['null', canonicalize_type(type_obj[:-1])]
        if len(type_obj) >= 2 and type_obj[-2:] == '[]':
            return {'type': 'array', 'items': canonicalize_type(type_obj[:-2])}
    if isinstance(type_obj, Dict):
        if type_obj.get('type') == 'array':
            return {**type_obj, 'items': canonicalize_type(type_obj['items'])}
    return type_obj


def canonicalize_steps_list(steps: Yaml) -> List[Yaml]:
    if isinstance(steps, list):
        all_dicts = all([isinstance(elt, dict) for elt in steps])
        if not all_dicts:
            msg = 'Error! If steps: tag is a List then all its elements should be Dictionaries!'
            raise Exception(f"{msg}\n{yaml.dump(steps)}")
        return steps
    if isinstance(steps, dict):
        items = [(key, {}) if val is None else (key, val) for key, val in steps.items()]
        all_dicts = all([isinstance(val, dict) for key, val in items])
        if not all_dicts:
            msg = 'Error! If steps: tag is a Dictionary then all its values should be Dictionaries!'
            raise Exception(f"{msg}\n{yaml.dump(steps)}")
        return [{'id': key, **val} for key, val in items]
    # steps should either be a list or a dict, but...
    return steps


def remove_id_tags(list_of_dicts_with_id_keys: list) -> dict[str, Yaml]:
    d_canon = {}
    for d in list_of_dicts_with_id_keys:
        id_tag = d['id']
        del d['id']
        # NOTE: The order of the steps may not be preserved!
        d_canon[id_tag] = d
    return d_canon


def canonicalize_steps_dict(steps: Yaml) -> Dict[str, Yaml]:
    if isinstance(steps, list):
        all_dicts = all([isinstance(elt, dict) for elt in steps])
        if not all_dicts:
            msg = 'Error! If steps: tag is a List then all its elements should be Dictionaries!'
            raise Exception(f"{msg}\n{yaml.dump(steps)}")
        return remove_id_tags(steps)
    if isinstance(steps, dict):
        all_dicts = all([isinstance(val, dict) for val in steps.values()])
        if not all_dicts:
            msg = 'Error! If steps: tag is a dictionary then all its values should be dictionaries!'
            raise Exception(f"{msg}\n{yaml.dump(steps)}")
        return steps
    # steps should either be a list or a dict, but...
    return steps


def canonicalize_inputs_dict(inputs: Yaml) -> Dict[str, Yaml]:
    inputs_canon = {}
    if isinstance(inputs, dict):
        for key, val in inputs.items():
            if isinstance(val, dict):
                inputs_canon[key] = val
            elif isinstance(val, str):
                inputs_canon[key] = {'type': val}  # NOTICE
            else:
                msg = 'Error! If inputs: tag is a dictionary, then all its values should be either strings (representing types) or dictionaries.'
                raise Exception(f"{msg}\n{yaml.dump(inputs)}")
    if isinstance(inputs, list):
        all_dicts = all([isinstance(elt, dict) for elt in inputs])
        if not all_dicts:
            msg = 'Error! If inputs: tag is a list then all its elements should be dictionaries!'
            raise Exception(f"{msg}\n{yaml.dump(inputs)}")
        return remove_id_tags(inputs)
    return inputs_canon


def canonicalize_outputs_dict(outputs: Yaml) -> Dict[str, Yaml]:
    outputs_canon = {}
    if isinstance(outputs, dict):
        for key, val in outputs.items():
            if isinstance(val, dict):
                outputs_canon[key] = val
            elif isinstance(val, str):
                # TODO need to lookup output file mapping!
                outputs_canon[key] = val  # type: ignore
            else:
                msg = 'Error! outputs: tag should be a dictionary whose values are either strings (representing output files) or dictionaries.'
                raise Exception(f"{msg}\n{yaml.dump(outputs)}")
    if isinstance(outputs, list):
        all_dicts = all([isinstance(elt, dict) for elt in outputs])
        if not all_dicts:
            msg = 'Error! If outputs: tag is a list then all its elements should be dictionaries!'
            raise Exception(f"{msg}\n{yaml.dump(outputs)}")
        return remove_id_tags(outputs)
    return outputs_canon


def desugar_into_canonical_normal_form(cwl: Yaml) -> Yaml:
    if 'inputs' in cwl:
        # Arbitrarily choose dict form
        cwl['inputs'] = canonicalize_inputs_dict(cwl['inputs'])
    if 'outputs' in cwl:
        cwl['outputs'] = canonicalize_outputs_dict(cwl['outputs'])
    if 'steps' in cwl:
        # NOTE: No steps: to canonicalize for class: CommandLineTool
        # Choose list form due to
        # 1. dict keys must be unique (thus cannot use the same CLT twice)
        # 2. Some AST transformations (i.e. python_script) need to mutate the step id in-place
        # (which is not possible in dict form)
        # 3. Inlineing a subworkflow dict into the parent workflow dict may also cause key collisions.
        cwl['steps'] = canonicalize_steps_list(cwl['steps'])
    return cwl


def copy_cwl_input_output_dict(io_dict: Dict, remove_qmark: bool = False) -> Dict:
    """Copies the type, format, label, and doc entries. Does NOT copy inputBinding and outputBinding.

    Args:
        io_dict (Dict): A dictionary
        remove_qmark (bool): Determines whether to remove question marks and thus make optional types required

    Returns:
        Dict: A copy of the dictionary.
    """
    io_type = io_dict['type']
    if isinstance(io_type, str) and remove_qmark:
        io_type = io_type.replace('?', '')  # Providing optional arguments makes them required
    new_dict = {'type': canonicalize_type(io_type)}
    for key in ['format', 'label', 'doc']:
        if key in io_dict:
            new_dict[key] = io_dict[key]  # copy.deepcopy() ?
    return new_dict



from datetime import timedelta
from pathlib import Path
import unittest

import graphviz
from hypothesis import given, settings, HealthCheck
import networkx as nx
import pytest

import sophios
import sophios.ast
import sophios.cli
import sophios.main
import sophios.plugins
import sophios.schemas
import sophios.schemas.wic_schema
import sophios.utils
from sophios.wic_types import GraphData, GraphReps, Yaml, YamlTree, StepId

from .test_setup import tools_cwl, yml_paths, validator, wic_strategy


@pytest.mark.skip_pypi_ci
class TestFuzzyCompile(unittest.TestCase):

    @pytest.mark.slow
    @given(wic_strategy)
    @settings(max_examples=100,
              suppress_health_check=[HealthCheck.too_slow,
                                     HealthCheck.filter_too_much],
              deadline=timedelta(milliseconds=20000))
    # TODO: Improve schema so we can remove the health checks
    def test_fuzzy_compile(self, yml: Yaml) -> None:
        """Tests that the compiler doesn't crash when given random allegedly valid input.\n
        Note that the full schema has performance limitations, so a random subset of\n
        wic_main_schema is chosen when hypothesis=True, then random values are generated.

        Args:
            yml (Yaml): Yaml input, randomly generated according to a random subset of wic_main_schema
        """
        plugin_ns = 'global'
        yml_path = Path('random_stepid')
        steps_keys = sophios.utils.get_steps_keys(yml.get('steps', []))
        subkeys = sophios.utils.get_subkeys(steps_keys)
        if subkeys:
            # NOTE: Since all filepaths are currently relative w.r.t. --yaml,
            # we need to supply a fake --yaml. Using [0] works because we are
            # using k=1 in wic_main_schema.
            yml_path_stem = Path(subkeys[0]).stem
            yml_path = yml_paths[plugin_ns][yml_path_stem]

        args = sophios.cli.get_args(str(yml_path))

        y_t = YamlTree(StepId('random_stepid', plugin_ns), yml)
        yaml_tree_raw = sophios.ast.read_ast_from_disk(args.homedir, y_t, yml_paths, tools_cwl, validator,
                                                       args.ignore_validation_errors)
        yaml_tree = sophios.ast.merge_yml_trees(yaml_tree_raw, {}, tools_cwl)
        root_yml_dir_abs = yml_path.parent.absolute()
        yaml_tree = sophios.ast.python_script_generate_cwl(yaml_tree, root_yml_dir_abs, tools_cwl)

        graph_gv = graphviz.Digraph(name=f'cluster_{yml_path}')
        graph_gv.attr(newrank='True')
        graph_nx = nx.DiGraph()
        graphdata = GraphData(str(yml_path))
        graph = GraphReps(graph_gv, graph_nx, graphdata)
        try:
            compiler_info = sophios.compiler.compile_workflow(yaml_tree, args, [], [graph], {}, {}, {},
                                                              {}, tools_cwl, True, relative_run_path=True,
                                                              testing=True)
        except Exception as e:
            multi_def_str = 'Error! Multiple definitions of &'
            unbound_lit_var = 'Error! Unbound literal variable ~'
            python_script = 'Error! Cannot load python_script'
            self_reference = 'Error! Cannot self-reference the same step!'
            # Certain constraints are conditionally dependent on values and are
            # not easily encoded in the schema, so catch them here.
            # Moreover, although we check for the existence of input files in
            # stage_input_files, we cannot encode file existence in json schema
            # to check the python_script script: tag before compile time.
            if multi_def_str in str(e) or unbound_lit_var in str(e) or python_script in str(e) or self_reference in str(e):
                pass
            else:
                # import yaml
                # print(yaml.dump(yml))
                raise e


if __name__ == '__main__':
    sophios.plugins.logging_filters()
    unittest.main()



import uuid
import copy
from pathlib import Path
import sys
import traceback
from typing import Dict, List

from mergedeep import merge, Strategy
from jsonschema import Draft202012Validator
import yaml

from . import python_cwl_adapter, utils, utils_cwl
from .wic_types import Yaml, Tools, YamlTree, YamlForest, StepId, Tool
from sophios.utils_yaml import wic_loader

# NOTE: AST = Abstract Syntax Tree


def read_ast_from_disk(homedir: str,
                       yaml_tree_tuple: YamlTree,
                       yml_paths: Dict[str, Dict[str, Path]],
                       tools: Tools,
                       validator: Draft202012Validator,
                       ignore_validation_errors: bool) -> YamlTree:
    """Reads the yml workflow definition files from disk (recursively) and inlines them into an AST

    Args:
        homedir (str): The users home directory
        yaml_tree_tuple (YamlTree): A tuple of a filepath and its Yaml file contents.
        yml_paths (Dict[str, Dict[str, Path]]): The yml workflow definitions found using get_yml_paths()
        tools (Tools): The CWL CommandLineTool definitions found using get_tools_cwl()
        validator (Draft202012Validator): Used to validate the yml files against the autogenerated schema.
        ignore_validation_errors (bool): Temporarily ignore validation errors. Do not use this permanently!

    Raises:
        Exception: If the yml file(s) do not exist

    Returns:
        YamlTree: A tuple of the root filepath and the associated yml AST
    """
    (step_id, yaml_tree) = yaml_tree_tuple

    yaml_tree = utils_cwl.desugar_into_canonical_normal_form(yaml_tree)

    try:
        if not ignore_validation_errors:
            validator.validate(yaml_tree)
    except Exception as e:
        yaml_path = Path(step_id.stem)
        print('Failed to validate', yaml_path)
        print(f'See validation_{yaml_path.stem}.txt for detailed technical information.')
        # Do not display a nasty stack trace to the user; hide it in a file.
        with open(f'validation_{yaml_path.stem}.txt', mode='w', encoding='utf-8') as f:
            # https://mypy.readthedocs.io/en/stable/common_issues.html#python-version-and-system-platform-checks
            if sys.version_info >= (3, 10):
                traceback.print_exception(type(e), value=e, tb=None, file=f)
            else:
                traceback.print_exception(etype=type(e), value=e, tb=None, file=f)
        sys.exit(1)

    wic = {'wic': yaml_tree.get('wic', {})}
    if 'implementations' in wic['wic']:
        # Recursively expand each implementation, but do NOT choose a specific implementation.
        # Require back_name to be .wic? For now, yes.
        implementations_trees = []
        for back_name, back in wic['wic']['implementations'].items():
            plugin_ns = wic['wic'].get('namespace', 'global')
            stepid = StepId(back_name, plugin_ns)
            implementations_tree = read_ast_from_disk(homedir, YamlTree(stepid, back), yml_paths, tools, validator,
                                                      ignore_validation_errors)
            implementations_trees.append(implementations_tree)
        yaml_tree['wic']['implementations'] = dict(implementations_trees)
        return YamlTree(step_id, yaml_tree)

    steps: List[Yaml] = yaml_tree['steps']
    wic_steps = wic['wic'].get('steps', {})
    steps_keys = utils.get_steps_keys(steps)
    subkeys = utils.get_subkeys(steps_keys)

    for i, step_key in enumerate(steps_keys):
        stem = Path(step_key).stem

        # Recursively read subworkflows, adding yml file contents
        if step_key in subkeys:
            # Check for namespaceing; otherwise use the namespace 'global'.
            # NOTE: For now, do not support overloading / parameter passing for
            # namespaces, because we would have to call merge_yml_trees here.
            # It could (easily?) be done, but right now we have excellent
            # separation of concerns between simply reading yml files from disk
            # and then performing AST transformations in-memory.
            sub_wic = wic_steps.get(f'({i+1}, {step_key})', {})
            plugin_ns = sub_wic.get('wic', {}).get('namespace', 'global')

            paths_ns_i = yml_paths.get(plugin_ns, {})
            if paths_ns_i == {}:
                wicdir = Path(homedir) / 'wic'
                raise Exception(
                    f"Error! namespace {plugin_ns} not found in yaml paths. Check 'search_paths_wic' in your config file")
            if stem not in paths_ns_i:
                msg = f'Error! {stem} not found in namespace {plugin_ns} when attempting to read {step_id.stem}.wic'
                if stem == 'in':
                    msg += f'\n(Check that you have properly indented the `in` tag in {step_id.stem})'
                raise Exception(msg)
            yaml_path = paths_ns_i[stem]

            if not (yaml_path.exists() and yaml_path.suffix == '.wic'):
                raise Exception(f'Error! {yaml_path} does not exist or is not a .wic file.')

            # Load the high-level yaml sub workflow file.
            with open(yaml_path, mode='r', encoding='utf-8') as y:
                sub_yaml_tree_raw: Yaml = yaml.load(y.read(), Loader=wic_loader())

            y_t = YamlTree(StepId(step_key, plugin_ns), sub_yaml_tree_raw)
            (step_id_, sub_yml_tree) = read_ast_from_disk(homedir, y_t, yml_paths, tools, validator,
                                                          ignore_validation_errors)

            steps_i_copy = {**steps[i]}
            step_i_id = steps[i]['id']
            del steps_i_copy['id']
            # Do not merge these two dicts; use subtree and parentargs so we can
            # apply subtree before compilation and parentargs after compilation.
            steps[i] = {'id': step_i_id, 'subtree': sub_yml_tree, 'parentargs': steps_i_copy}

    return YamlTree(step_id, yaml_tree)


def merge_yml_trees(yaml_tree_tuple: YamlTree,
                    wic_parent: Yaml,
                    tools: Tools) -> YamlTree:
    """Implements 'parameter passing' by recursively merging wic: yml tags.
    Values from the parent workflow will overwrite / override subworkflows.
    See https://github.com/PolusAI/mm-workflows/blob/main/examples/gromacs/basic.wic for details

    Args:
        yaml_tree_tuple (YamlTree): A tuple of a name and a yml AST
        wic_parent (Yaml): The wic: yml dict from the parent workflow
        tools (Tools): The CWL CommandLineTool definitions found using get_tools_cwl()

    Raises:
        Exception: If a wic: tag is found as an argument to a CWL CommandLineTool

    Returns:
        YamlTree: The yml AST with all wic: tags recursively merged.
    """
    (step_id, yaml_tree) = yaml_tree_tuple

    # Check for top-level yml dsl args
    wic_self = {'wic': yaml_tree.get('wic', {})}
    wic = merge(wic_self, wic_parent, strategy=Strategy.TYPESAFE_REPLACE)
    # Here we want to ADD wic: as a top-level yaml tag.
    # In the compilation phase, we want to remove it.
    yaml_tree['wic'] = wic['wic']
    wic_steps = wic['wic'].get('steps', {})

    if 'implementations' in wic['wic']:
        # Recursively expand each implementation, but do NOT choose a specific implementation.
        # Require back_name to be .wic? For now, yes.
        implementations_trees = []
        for stepid, back in wic['wic']['implementations'].items():
            implementations_tree = merge_yml_trees(YamlTree(stepid, back), wic_parent, tools)
            implementations_trees.append(implementations_tree)
        yaml_tree['wic']['implementations'] = dict(implementations_trees)
        return YamlTree(step_id, yaml_tree)

    steps: List[Yaml] = yaml_tree['steps']
    steps_keys = utils.get_steps_keys(steps)
    subkeys = utils.get_subkeys(steps_keys)

    for i, step_key in enumerate(steps_keys):
        # Recursively merge subworkflows, to implement parameter passing.
        if step_key in subkeys:
            # Extract the sub yaml file that we pre-loaded from disk.
            sub_yml_tree_initial = steps[i]['subtree']
            sub_wic = wic_steps.get(f'({i+1}, {step_key})', {})

            y_t = YamlTree(StepId(step_key, step_id.plugin_ns), sub_yml_tree_initial)
            (step_key_, sub_yml_tree) = merge_yml_trees(y_t, sub_wic, tools)
            # Now mutably overwrite the self args with the merged args
            steps[i]['subtree'] = sub_yml_tree

        # Extract provided CWL args, if any, and (recursively) merge them with
        # provided CWL args passed in from the parent, if any.
        # (At this point, any DSL args provided from the parent(s) should have
        # all of the initial yml tags removed, leaving only CWL tags remaining.)
        if step_key not in subkeys:
            clt_args = wic_steps.get(f'({i+1}, {step_key})', {})
            if 'wic' in clt_args:
                # Do NOT add yml tags to the raw CWL!
                # We can simply leave any step-specific wic: tags at top-level.
                # Copy so we only delete from the step, not also the top-level.
                clt_args = copy.deepcopy(clt_args)
                del clt_args['wic']
            sub_yml_tree = clt_args
            args_provided_dict_self = steps[i]
            steps_i_id = steps[i]['id']
            # NOTE: To support overloading, the parent args must overwrite the child args!
            args_provided_dict = merge(args_provided_dict_self, sub_yml_tree,
                                       strategy=Strategy.TYPESAFE_REPLACE)  # TYPESAFE_ADDITIVE ?
            # Now mutably overwrite the self args with the merged args
            steps[i] = {**args_provided_dict, 'id': steps_i_id}

    return YamlTree(step_id, yaml_tree)


def tree_to_forest(yaml_tree_tuple: YamlTree, tools: Tools) -> YamlForest:
    """The purpose of this function is to abstract away the process of traversing an AST.

    Args:
        yaml_tree_tuple (YamlTree): A tuple of name and yml AST
        tools (Tools): The CWL CommandLineTool definitions found using get_tools_cwl()

    Returns:
        YamlForest: A recursive data structure containing all sub-trees encountered while traversing the yml AST.
    """
    (step_id, yaml_tree) = yaml_tree_tuple

    wic = {'wic': yaml_tree.get('wic', {})}
    if 'implementations' in wic['wic']:
        implementations_forest_list = []
        for stepid, back in wic['wic']['implementations'].items():
            implementation_forest = (stepid, tree_to_forest(YamlTree(stepid, back), tools))
            implementations_forest_list.append(implementation_forest)
        return YamlForest(YamlTree(step_id, yaml_tree), implementations_forest_list)

    steps: List[Yaml] = yaml_tree['steps']
    wic_steps = wic['wic'].get('steps', {})
    steps_keys = utils.get_steps_keys(steps)
    subkeys = utils.get_subkeys(steps_keys)

    yaml_forest_list = []

    for i, step_key in enumerate(steps_keys):

        if step_key in subkeys:
            wic_step_i = wic_steps.get(f'({i+1}, {step_key})', {})
            plugin_ns_i = wic_step_i.get('wic', {}).get('namespace', 'global')

            sub_yaml_tree = steps[i]['subtree']
            sub_yml_forest = tree_to_forest(YamlTree(StepId(step_key, plugin_ns_i), sub_yaml_tree), tools)
            (sub_yml_tree_step_id, sub_yml_tree_) = sub_yml_forest.yaml_tree
            yaml_forest_list.append((sub_yml_tree_step_id, sub_yml_forest))

    return YamlForest(YamlTree(step_id, yaml_tree), yaml_forest_list)


def python_script_generate_cwl(yaml_tree_tuple: YamlTree,
                               root_yml_dir_abs: Path,
                               tools: Tools) -> YamlTree:
    """Generates a CWL CommandLineTool for each python_script: tag,
    mutably adds them to tools, and updates the call sites in yaml_tree.

    Args:
        yaml_tree_tuple (YamlTree): A tuple of a name and a yml AST
        root_yml_dir_abs (Path): The absolute path to the directory containing the root workflow yml file
        tools (Tools): The CWL CommandLineTool definitions found using get_tools_cwl()

    Returns:
        YamlTree: The yml AST with all python_script tags replaced with references to the auto-generated CWL.
    """
    (step_id, yaml_tree) = yaml_tree_tuple

    wic = {'wic': yaml_tree.get('wic', {})}

    if 'implementations' in wic['wic']:
        implementations_trees = []
        for stepid, back in wic['wic']['implementations'].items():
            implementations_tree = python_script_generate_cwl(YamlTree(stepid, back), root_yml_dir_abs, tools)
            implementations_trees.append(implementations_tree)
        yaml_tree['wic']['implementations'] = dict(implementations_trees)
        return YamlTree(step_id, yaml_tree)

    steps: List[Yaml] = yaml_tree['steps']
    steps_keys = utils.get_steps_keys(steps)
    subkeys = utils.get_subkeys(steps_keys)

    for i, step_key in enumerate(steps_keys):
        if step_key in subkeys:
            sub_yml_tree_initial = steps[i]['subtree']
            y_t = YamlTree(StepId(step_key, step_id.plugin_ns), sub_yml_tree_initial)
            (step_key_, sub_yml_tree) = python_script_generate_cwl(y_t, root_yml_dir_abs, tools)
            steps[i]['subtree'] = sub_yml_tree

        if step_key not in subkeys:
            if 'python_script' == step_key:
                # This generates a CWL CommandLineTool for an arbitrary python script.
                yml_args = copy.deepcopy(steps[i]['in'])
                python_script_path = yml_args.get('script', '')
                if isinstance(python_script_path, dict) and 'wic_inline_input' in python_script_path:
                    python_script_path = python_script_path['wic_inline_input']
                # NOTE: The existence of the script: tag should now be guaranteed in the schema
                del yml_args['script']
                python_script_docker_pull = yml_args.get('dockerPull', '')  # Optional
                if isinstance(python_script_docker_pull, dict) and 'wic_inline_input' in python_script_docker_pull:
                    python_script_docker_pull = python_script_docker_pull['wic_inline_input']
                if 'dockerPull' in yml_args:
                    del yml_args['dockerPull']
                    del steps[i]['in']['dockerPull']
                python_script_path = root_yml_dir_abs / Path(python_script_path)
                python_script_mod = Path(python_script_path).name[:-3]
                module = python_cwl_adapter.get_module(python_script_mod, python_script_path, yml_args)
                generated_cwl = python_cwl_adapter.generate_CWL_CommandLineTool(
                    module.inputs, module.outputs, python_script_docker_pull)
                # See https://docs.python.org/3/library/uuid.html#uuid.uuid4
                # "If all you want is a unique ID, you should probably call uuid1() or uuid4()"
                unique_id = 'python_script_' + str(uuid.uuid4())
                # TODO: use the trick from the copy_cwl_tools branch to keep
                # programmatically modified (and in this case, generated from scratch)
                # CWL CommandLineTools in-memory so we can defer writing to disk
                filepath = 'autogenerated/' + unique_id + '.cwl'
                with open(filepath, mode='w', encoding='utf-8') as f:
                    f.write(yaml.dump(generated_cwl, sort_keys=False, line_break='\n', indent=2))

                # Now replace step_key with unique_id in the workflow
                # NOTE: In addition to duplicate keys,
                # this is why we are forced to use canonicalize_steps_list
                # (We can replace a list element in-place, but if we delete a
                # dict key and add another key, the second key gets added to the
                # end of the dict...)
                steps[i] = {**steps[i], 'id': unique_id}
                # and add the auto-generated CWL CommandLineTool to tools
                step_id_ = StepId(unique_id, 'global')
                tool_i = Tool(filepath, generated_cwl)
                tools[step_id_] = tool_i

    return YamlTree(step_id, yaml_tree)




from . import _version
__version__ = _version.get_versions()['version']

# Well, not THIS file.
auto_gen_header = f"""# This file was autogenerated using the Workflow Inference Compiler, version {__version__}
# https://github.com/PolusAI/workflow-inference-compiler\n"""



import argparse
import copy
import json
import os
from pathlib import Path
import sys
from typing import Dict, List

import graphviz
from mergedeep import merge, Strategy
import networkx as nx
import yaml


from . import input_output as io
from . import inference, utils, utils_cwl, utils_graphs
from .wic_types import (CompilerInfo, EnvData, ExplicitEdgeCalls,
                        ExplicitEdgeDefs, GraphData, GraphReps, Namespaces,
                        NodeData, RoseTree, Tool, Tools, WorkflowInputsFile,
                        Yaml, YamlTree, StepId)

# NOTE: This must be initialized in main.py and/or cwl_subinterpreter.py
inference_rules: Dict[str, str] = {}


def compile_workflow(yaml_tree_ast: YamlTree,
                     args: argparse.Namespace,
                     namespaces: Namespaces,
                     subgraphs_: List[GraphReps],
                     explicit_edge_defs: ExplicitEdgeDefs,
                     explicit_edge_calls: ExplicitEdgeCalls,
                     input_mapping: Dict[str, List[str]],
                     output_mapping: Dict[str, str],
                     tools: Tools,
                     is_root: bool,
                     relative_run_path: bool,
                     testing: bool) -> CompilerInfo:
    """fixed-point wrapper around compile_workflow_once\n
    See https://en.wikipedia.org/wiki/Fixed_point_(mathematics)

    Args:
        yaml_tree_ast (YamlTree): A tuple of name and yml AST
        args (Any): all of the other positional arguments for compile_workflow_once
        kwargs (Any): all of the other keyword arguments for compile_workflow_once

    Returns:
        CompilerInfo: Contains the data associated with compiled subworkflows\n
        (in the Rose Tree) together with mutable cumulative environment\n
        information which needs to be passed through the recursion.
    """
    ast_modified = True
    yaml_tree = yaml_tree_ast
    # There ought to be at most one insertion between each step.
    # If everything is working correctly, we should thus reach the fixed point
    # in at most n-1 iterations. However, due to the possibility of bugs in the
    # implementation and/or spurious inputs, we should guarantee termination.
    max_iters = 100  # 100 ought to be plenty. TODO: calculate n-1 from steps:
    i = 0
    while ast_modified and i < max_iters:
        subgraphs = copy.deepcopy(subgraphs_)  # See comment below!
        compiler_info = compile_workflow_once(yaml_tree, args, namespaces, subgraphs,
                                              explicit_edge_defs, explicit_edge_calls,
                                              input_mapping, output_mapping,
                                              tools, is_root, relative_run_path, testing)
        node_data: NodeData = compiler_info.rose.data
        ast_modified = not yaml_tree.yml == node_data.yml
        if ast_modified:
            # import yaml
            # print(yaml.dump(node_data.yml))
            # print()
            yaml_tree = YamlTree(yaml_tree_ast.step_id, node_data.yml)
        i += 1

    # Overwrite subgraphs_ element-wise
    # This is a terrible hack due to the fact that the graphviz library API
    # only allows appending to the body. This introduces mutable state, so each
    # time we speculatively compile we accumulate duplicate nodes and edges.
    # The 'correct' solution is to store the nodes and edges that you want to
    # add in a separate data structure, return them from compile_workflow_once,
    # and only once we have reached the fixed point then add them here. Due to
    # labeling and styling and other graphviz metadata that is not trivial, so
    # instead we simply deepcopy and overwrite the bodies here.
    # (NOTE: We now have a separate GraphData structure, so the graphviz and
    # networkx representations can probably be removed from the recursion.)
    # (Also note that you have to do this element-wise; you cannot simply write
    # subgraphs_ = subgraphs because that will only overwrite the local binding
    # and thus it will not affect the call site of compile_workflow!)
    for i, subgraph_ in enumerate(subgraphs_):
        subgraph_.graphviz.body = subgraphs[i].graphviz.body
        subgraph_.graphdata.name = subgraphs[i].graphdata.name
        subgraph_.graphdata.nodes = subgraphs[i].graphdata.nodes
        subgraph_.graphdata.edges = subgraphs[i].graphdata.edges
        subgraph_.graphdata.subgraphs = subgraphs[i].graphdata.subgraphs
        subgraph_.networkx.clear()
        subgraph_.networkx.update(subgraphs[i].networkx.edges, subgraphs[i].networkx.nodes)

    if i == max_iters:
        import yaml
        print(yaml.dump(node_data.yml))
        raise Exception(f'Error! Maximum number of iterations ({max_iters}) reached in compile_workflow!')
    return compiler_info


def compile_workflow_once(yaml_tree_ast: YamlTree,
                          args: argparse.Namespace,
                          namespaces: Namespaces,
                          subgraphs: List[GraphReps],
                          explicit_edge_defs: ExplicitEdgeDefs,
                          explicit_edge_calls: ExplicitEdgeCalls,
                          input_mapping: Dict[str, List[str]],
                          output_mapping: Dict[str, str],
                          tools: Tools,
                          is_root: bool,
                          relative_run_path: bool,
                          testing: bool) -> CompilerInfo:
    """STOP: Have you read the Developer's Guide?? docs/devguide.md\n
    Recursively compiles yml workflow definition ASTs to CWL file contents

    Args:
        yaml_tree_ast (YamlTree): A tuple of name and yml AST
        args (argparse.Namespace): The command line arguments
        namespaces (Namespaces): Specifies the path in the yml AST to the current subworkflow
        subgraphs (List[Graph]): The graphs associated with the parent workflows of the current subworkflow
        explicit_edge_defs (ExplicitEdgeDefs): Stores the (path, value) of the explicit edge definition sites
        explicit_edge_calls (ExplicitEdgeCalls): Stores the (path, value) of the explicit edge call sites
        tools (Tools): The CWL CommandLineTool definitions found using get_tools_cwl().\n
        yml files that have been compiled to CWL SubWorkflows are also added during compilation.
        is_root (bool): True if this is the root workflow
        relative_run_path (bool): Controls whether to use subdirectories or\n
        just one directory when writing the compiled CWL files to disk
        testing: Used to disable some optional features which are unnecessary for testing.

    Raises:
        Exception: If any errors occur

    Returns:
        CompilerInfo: Contains the data associated with compiled subworkflows\n
        (in the Rose Tree) together with mutable cumulative environment\n
        information which needs to be passed through the recursion.
    """
    # NOTE: Use deepcopy so that when we delete wic: we don't modify any call sites
    (step_id, yaml_tree) = copy.deepcopy(yaml_tree_ast)
    yaml_path = step_id.stem
    # We also want another copy of the original AST so that if we need to modify it,
    # we can return the modified AST to the call site and re-compile.
    (yaml_path_orig, yaml_tree_orig) = copy.deepcopy(yaml_tree_ast)

    if not testing:
        print(' starting', ('  ' * len(namespaces)) + yaml_path)

    # Check for top-level yml dsl args
    wic = {'wic': yaml_tree.get('wic', {})}
    # import yaml; print(yaml.dump(wic))
    if 'wic' in yaml_tree:
        del yaml_tree['wic']
    wic_steps = wic['wic'].get('steps', {})

    yaml_stem = Path(yaml_path).stem

    (back_name_, yaml_tree) = utils.extract_implementation(yaml_tree, wic['wic'], Path(yaml_path))
    steps: List[Yaml] = yaml_tree['steps']

    steps_keys = utils.get_steps_keys(steps)

    subkeys = utils.get_subkeys(steps_keys)

    # Add headers
    # Use 1.0 because cromwell only supports 1.0 and we are not using 1.1 / 1.2 features.
    # Eventually we will want to use 1.2 to support conditional workflows and 1.3 to support loops.
    yaml_tree['cwlVersion'] = yaml_tree.get('cwlVersion', 'v1.2')
    yaml_tree['class'] = 'Workflow'
    yaml_tree['$namespaces'] = {**yaml_tree.get('$namespaces', {}), **{'edam': 'https://edamontology.org/'}}
    edam_dev_owl = 'https://raw.githubusercontent.com/edamontology/edamontology/master/EDAM_dev.owl'
    schemas = yaml_tree.get('$schemas', [])
    if edam_dev_owl not in schemas:
        schemas.append(edam_dev_owl)
        yaml_tree['$schemas'] = schemas

    # Collect workflow input parameters
    inputs_workflow = {}
    inputs_file_workflow = {}

    # Collect workflow input/output to workflow step input/output mappings
    input_mapping_copy = copy.deepcopy(input_mapping)
    output_mapping_copy = copy.deepcopy(output_mapping)

    # Collect the internal workflow output variables
    outputs_workflow = []
    vars_workflow_output_internal = []

    # Copy recursive explicit edge variable definitions and call sites.
    explicit_edge_defs_copy = copy.deepcopy(explicit_edge_defs)
    explicit_edge_calls_copy = copy.deepcopy(explicit_edge_calls)
    # Unlike the first copies which are mutably updated, these are returned
    # unmodified so that we can test that compilation is embedding independent.
    explicit_edge_defs_copy2 = copy.deepcopy(explicit_edge_defs)
    explicit_edge_calls_copy2 = copy.deepcopy(explicit_edge_calls)

    # Collect recursive subworkflow data
    step_1_names = []
    sibling_subgraphs = []

    rose_tree_list = []

    graph = subgraphs[-1]  # Get the current graph
    graph_gv = graph.graphviz
    graph_nx = graph.networkx
    graphdata = graph.graphdata

    # plugin_ns = wic['wic'].get('namespace', 'global')

    tools_lst: List[Tool] = []

    for i, step_key in enumerate(steps_keys):
        step_name_i = utils.step_name_str(yaml_stem, i, step_key)
        stem = Path(step_key).stem
        wic_step_i = wic_steps.get(f'({i+1}, {step_key})', {})
        # NOTE: See comments in read_ast_from_disk()
        plugin_ns_i = wic_step_i.get('wic', {}).get('namespace', 'global')  # plugin_ns ??
        stepid = StepId(stem, plugin_ns_i)

        step_name_or_key = step_name_i if step_key.endswith('.wic') or stepid in tools else step_key

        # Recursively compile subworkflows, adding compiled cwl file contents to tools
        ast_modified = False
        if step_key in subkeys:
            # NOTE: step_name_or_key should always be step_name_i in this branch
            # Extract the sub yaml file that we pre-loaded from disk.
            sub_yml = steps[i]['subtree']
            sub_yaml_tree = YamlTree(StepId(step_key, plugin_ns_i), sub_yml)

            # get the label (if any) from the subworkflow
            step_i_wic_graphviz = sub_yml.get('wic', {}).get('graphviz', {})
            label = step_i_wic_graphviz.get('label', step_key)
            style = step_i_wic_graphviz.get('style', '')

            subgraph_gv = graphviz.Digraph(name=f'cluster_{step_key}')
            subgraph_gv.attr(label=label)  # str(path)
            subgraph_gv.attr(color='lightblue')  # color of outline
            if style != '':
                subgraph_gv.attr(style=style)
            subgraph_nx = nx.DiGraph()
            graphdata = GraphData(step_key)
            subgraph = GraphReps(subgraph_gv, subgraph_nx, graphdata)

            sub_compiler_info = compile_workflow(sub_yaml_tree, args, namespaces + [step_name_or_key],
                                                 subgraphs + [subgraph], explicit_edge_defs_copy,
                                                 explicit_edge_calls_copy,
                                                 input_mapping_copy, output_mapping_copy,
                                                 tools, False, relative_run_path, testing)

            sub_rose_tree = sub_compiler_info.rose
            rose_tree_list.append(sub_rose_tree)

            sub_node_data: NodeData = sub_rose_tree.data
            sub_env_data = sub_compiler_info.env

            ast_modified = not sub_yaml_tree.yml == sub_node_data.yml
            if ast_modified:
                # Propagate the updated yaml_tree (and wic: tags) upwards.
                # Since we already called ast.merge_yml_trees() before initially
                # compiling, the only way the child wic: tags can differ from
                # the parent is if there were modifications during compilation.
                # In other words, it should be safe to simply replace the
                # relevant yaml_tree and wic: tags in the parent with the child
                # values.
                print('AST modified', step_key)
                wic_steps[f'({i+1}, {step_key})'] = {'wic': sub_node_data.yml.get('wic', {})}
                wic_step_i = wic_steps.get(f'({i+1}, {step_key})', {})
                # import yaml
                # print(yaml.dump(wic_steps))

            # Add arguments to the compiled subworkflow (if any), being careful
            # to remove any child wic: metadata annotations. Post-compilation
            # arguments can now be added either directly inline or as metadata.
            wic_step_i_copy = copy.deepcopy(wic_step_i)
            if 'wic' in wic_step_i_copy:
                del wic_step_i_copy['wic']
            # NOTE: To support overloading, the metadata args must overwrite the parent args!
            steps_i_id = steps[i]['id']
            args_provided_dict = merge(steps[i]['parentargs'], wic_step_i_copy,
                                       strategy=Strategy.TYPESAFE_REPLACE)  # TYPESAFE_ADDITIVE ?
            steps[i] = {**args_provided_dict, 'id': steps_i_id}

            sibling_subgraphs.append(sub_node_data.graph)  # TODO: Just subgraph?
            step_1_names.append(sub_node_data.step_name_1)
            tool_i = Tool(stem + '.cwl', sub_node_data.compiled_cwl)

            # Initialize the above from recursive values.
            # Do not initialize inputs_workflow. See comment below.
            # inputs_workflow.update(sub_node_data.inputs_workflow)
            inputs_namespaced_list = []
            for k, val in sub_env_data.inputs_file_workflow.items():
                input_namespaced = (f'{step_name_or_key}___{k}', val)  # _{step_key}_input___{k}
                inputs_namespaced_list.append(input_namespaced)
            inputs_namespaced = dict(inputs_namespaced_list)
            inputs_file_workflow.update(inputs_namespaced)

            input_mapping_copy_namespaced = dict([(f'{step_name_or_key}___{k}', val) for k, val in
                                                  sub_env_data.input_mapping.items() if k not in input_mapping])
            input_mapping_copy.update(input_mapping_copy_namespaced)

            output_mapping_copy_namespaced = dict([(f'{step_name_or_key}___{k}', val) for k, val in
                                                   sub_env_data.output_mapping.items() if k not in output_mapping])
            output_mapping_copy.update(output_mapping_copy_namespaced)

            vars_workflow_output_internal += sub_env_data.vars_workflow_output_internal
            explicit_edge_defs_copy.update(sub_env_data.explicit_edge_defs)
            explicit_edge_calls_copy.update(sub_env_data.explicit_edge_calls)
        else:
            run_tag = steps[i].get('run', '')
            run_tag_stem = Path(run_tag).stem if isinstance(run_tag, str) else ''
            stepid_runtag = StepId(run_tag_stem, 'global')

            if stepid in tools:
                # Use auto-discovery mechanism (with step_key)
                tool_i = tools[stepid]
            elif stepid_runtag in tools:
                # Use auto-discovery mechanism (with run tag)
                tool_i = tools[stepid_runtag]
            else:
                msg = f'Error! Neither {stepid.stem} nor {stepid_runtag.stem} found!'
                raise Exception(msg)
            # Programmatically modify tool_i here
            graph_dummy = graph  # Just use anything here to satisfy mypy
            name_stem = Path(tool_i.run_path).stem
            node_data_base_case = NodeData([step_name_or_key], name_stem, {}, tool_i.cwl,
                                           tool_i, {}, {}, {}, graph_dummy, {}, '')
            rose_tree_base_case = RoseTree(node_data_base_case, [])
            rose_tree_list.append(rose_tree_base_case)
        tools_lst.append(tool_i)

        if not testing:
            # Disable for testing because when testing in parallel, the *.gv Graphviz files
            # can be written/read to/from disk simultaneously, which results in
            # intermittent 'syntax errors'.
            pass
            # Actually, this is a significant performance bottleneck and isn't really necessary.
            # utils_graphs.make_tool_dag(stem, tool_i, args.graph_dark_theme)

        # Add run tag, using relative or flat-directory paths
        # NOTE: run: path issues were causing test_cwl_embedding_independence()
        # to fail, so I simply ignore the run tag in that test.
        # run_path = tool_i.run_path
        # Instead of referencing the original CWL CommandLineTools in-place,
        # let's copy them to autogenerated/ so we can programmatically modify them.
        # Note that this will cause name collisions if relative_run_path=False
        # and when using the 'same' CommandLineTool twice. i.e. two different
        # modifications will get written to the same cwl file, so the last one 'wins'.
        # (So don't use relative_run_path=False)
        # run_path = step_name_or_key + '/' + Path(tool_i.run_path).stem + '.cwl'
        run_path = Path(tool_i.run_path).stem + '.cwl'
        # NOTE: run_path is always relative; relative_run_path should probably
        # be called use_subdirs, because it simply determines if subworkflows
        # should be written to subdirectories or if everything should be
        # written to autogenerated/
        if relative_run_path:
            # if step_key in subkeys:
            #     run_path = step_name_or_key + '/' + run_path
            # else:
            #     run_path = os.path.relpath(run_path, 'autogenerated/')
            #     run_path = ('../' * len(namespaces)) + run_path
            run_path = step_name_or_key + '/' + run_path
        else:
            if step_key in subkeys:
                run_path = '___'.join(namespaces + [step_name_or_key, run_path])
            else:
                run_path = os.path.relpath(run_path, 'autogenerated/')

        if not 'run' in steps[i]:
            steps[i]['run'] = run_path
        elif isinstance(steps[i]['run'], str):
            steps[i]['run'] = run_path
        else:
            pass  # run: tag can also contain inlined file contents (currently unsupported)

        # Generate intermediate file names between steps.
        if 'in' not in steps[i]:
            steps[i]['in'] = {}

        if 'cwl_subinterpreter' == step_key:
            in_dict_in = steps[i]['in']  # NOTE: Mutates in_dict_in
            io.write_absolute_yaml_tags(args, in_dict_in, namespaces, step_name_or_key, explicit_edge_calls_copy)

        args_provided = []
        if 'in' in steps[i]:
            args_provided = list(steps[i]['in'])
        # print(args_provided)

        in_tool = tool_i.cwl['inputs']
        # print(list(in_tool.keys()))
        if tool_i.cwl['class'] == 'CommandLineTool':
            args_required = [arg for arg in in_tool if not (in_tool[arg].get('default') or
                                                            # Check for optional arguments using both the '?' syntactic sugar, as well as the
                                                            # canonical null representation. See canonicalize_type in cwl_utils.py
                                                            (isinstance(in_tool[arg]['type'], str) and in_tool[arg]['type'][-1] == '?') or
                                                            (isinstance(in_tool[arg]['type'], List) and 'null' in in_tool[arg]['type']))]
        elif tool_i.cwl['class'] == 'Workflow':
            args_required = list(in_tool)

            if 'in' not in steps[i]:
                steps[i]['in'] = {key: key for key in args_required}
            else:
                # Add keys, but do not overwrite existing vals.
                for key in args_required:
                    if key not in steps[i]['in']:
                        steps[i]['in'][key] = key
        else:
            raise Exception('Unknown class', tool_i.cwl['class'])

        # Note: Some biobb config tags are not required in the cwl files, but are in
        # fact required in the python source code! See check_mandatory_property
        # (Solution: refactor all required arguments out of config and list
        # them as explicit inputs in the cwl files, then modify the python
        # files accordingly.)
        # print(args_required)

        sub_args_provided = [arg for arg in args_required if arg in explicit_edge_calls_copy]
        # print(sub_args_provided)

        label = step_key
        if args.graph_label_stepname:
            label = step_name_or_key
        step_node_name = '___'.join(namespaces + [step_name_or_key])

        if not tool_i.cwl['class'] == 'Workflow':
            wic_graphviz_step_i = wic_step_i.get('wic', {}).get('graphviz', {})
            label = wic_graphviz_step_i.get('label', label)
            default_style = 'rounded, filled'
            style = wic_graphviz_step_i.get('style', '')
            style = default_style if style == '' else default_style + ', ' + style
            attrs = {'label': label, 'shape': 'box', 'style': style, 'fillcolor': 'lightblue'}
            graph_gv.node(step_node_name, **attrs)
            graph_nx.add_node(step_node_name)
            graphdata.nodes.append((step_node_name, attrs))
        elif not (step_key in subkeys and len(namespaces) < args.graph_inline_depth):
            nssnode = namespaces + [step_name_or_key]
            # Just like in add_graph_edge(), here we can hide all of the details
            # below a given depth by simply truncating the node's namespaces.
            nssnode = nssnode[:(1 + args.graph_inline_depth)]
            step_node_name = '___'.join(nssnode)
            # NOTE: NOT wic_graphviz_step_i
            # get the label (if any) from the subworkflow
            # TODO: This causes test_cwl_embedding_independence to fail.
            # yml = sub_node_data.yml if ast_modified else sub_yaml_tree.yml
            # step_i_wic_graphviz = yml.get('wic', {}).get('graphviz', {})
            # TODO: For insertions, figure out why this is using
            # the label from the parent workflow.
            # label = step_i_wic_graphviz.get('label', label)
            default_style = 'rounded, filled'
            style = ''  # step_i_wic_graphviz.get('style', '')
            style = default_style if style == '' else default_style + ', ' + style
            attrs = {'label': label, 'shape': 'box', 'style': style, 'fillcolor': 'lightblue'}
            graph_gv.node(step_node_name, **attrs)
            graph_nx.add_node(step_node_name)
            graphdata.nodes.append((step_node_name, attrs))

        if 'out' not in steps[i]:
            steps[i]['out'] = []

        if isinstance(steps[i]['out'], Dict):
            print('Error! The `out` tag should be a list, not a dictionary!')
        for j in range(len(steps[i]['out'])):
            out_val = steps[i]['out'][j]
            if isinstance(out_val, Dict):
                keys = list(out_val.keys())
                if len(keys) != 1:
                    print('Error! There should only be one anchor per out: list entry!')
                    # raise exception?
                out_key = keys[0]
                out_val = out_val[out_key]
                if isinstance(out_val, Dict) and 'wic_anchor' in out_val:
                    edgedef = out_val['wic_anchor']

                    # NOTE: There can only be one definition, but multiple call sites.
                    if not explicit_edge_defs_copy.get(edgedef):
                        steps[i]['out'][j] = out_key  # discard anchor / retain string key
                        explicit_edge_defs_copy.update({edgedef: (namespaces + [step_name_or_key], out_key)})
                        # Add a 'dummy' value to explicit_edge_calls, because
                        # that determines sub_args_provided when the recursion returns.
                        explicit_edge_calls_copy.update({edgedef: (namespaces + [step_name_or_key], out_key)})
                    else:
                        raise Exception(f"Error! Multiple definitions of &{edgedef}!")

        # NOTE: sub_args_provided are handled within the args_required loop below
        for arg_key in args_provided:
            # Extract input value into separate yml file
            # Replace it here with a new variable name
            arg_val = steps[i]['in'][arg_key]

            # Convert native YAML to a JSON-encoded string for specific tags.
            tags = ['config']
            if arg_key in tags and isinstance(arg_val, Dict) and ('wic_inline_input' in arg_val):
                arg_val = {'wic_inline_input': json.dumps(arg_val['wic_inline_input'])}

            # Use triple underscore for namespacing so we can split later
            in_name = f'{step_name_or_key}___{arg_key}'

            # Add auxiliary inputs for scatter steps
            if str(arg_key).startswith('__') and str(arg_key).endswith('__'):
                in_dict = {'type': arg_val['type']}
                inputs_workflow.update({in_name: in_dict})
                in_dict = {**in_dict, 'value': arg_val}
                inputs_file_workflow.update({in_name: in_dict})
                steps[i]['in'][arg_key] = {'source': in_name}
                continue

            in_dict = utils_cwl.copy_cwl_input_output_dict(in_tool[arg_key], True)

            if isinstance(arg_val, Dict) and 'wic_alias' in arg_val:
                arg_val = arg_val['wic_alias']
                if not explicit_edge_defs_copy.get(arg_val):
                    if is_root and not testing:
                        # Even if is_root, we don't want to raise an Exception
                        # here because in test_cwl_embedding_independence, we
                        # recompile all subworkflows as if they were root. That
                        # will cause this code path to be taken but it is not
                        # actually an error. Add a CWL input for testing only.
                        raise Exception(f"Error! No definition found for !&{arg_val}!")
                    inputs_workflow.update({in_name: in_dict})
                    steps[i]['in'][arg_key] = {'source': in_name}
                    # Add a 'dummy' value to explicit_edge_calls anyway, because
                    # that determines sub_args_provided when the recursion returns.
                    explicit_edge_calls_copy.update({in_name: (namespaces + [step_name_or_key], arg_key)})
                else:
                    (nss_def_init, var) = explicit_edge_defs_copy[arg_val]

                    nss_def_embedded = var.split('___')[:-1]
                    nss_call_embedded = arg_key.split('___')[:-1]
                    nss_def = nss_def_init + nss_def_embedded
                    # [step_name_or_key] is correct; nss_def_init already contains it from the recursive call
                    nss_call = namespaces + [step_name_or_key] + nss_call_embedded

                    nss_def_inits, nss_def_tails = utils.partition_by_lowest_common_ancestor(nss_def, nss_call)
                    nss_call_inits, nss_call_tails = utils.partition_by_lowest_common_ancestor(nss_call, nss_def)
                    # nss_def and nss_call are paths into the abstract syntax tree 'call stack'.
                    # This defines the 'common namespace' in the call stack w.r.t. the inits.
                    assert nss_def_inits == nss_call_inits

                    # TODO: Check this comment.
                    # Relative to the common namespace, if the call site of an explicit
                    # edge is at a depth > 1, (i.e. if it is NOT simply of the form
                    # last_namespace/input_variable) then we
                    # need to create inputs in all of the intervening CWL files
                    # so we can pass in the values from the outer scope(s). Here,
                    # we simply need to use in_name and add to inputs_workflow
                    # and explicit_edge_calls. The outer scope(s) are handled by
                    # the sub_args_provided clause below.
                    # Note that the outputs from the definition site are bubbled
                    # up the call stack until they reach the common namespace.
                    if len(nss_call_tails) == 1:
                        # TODO: Check this comment.
                        # The definition site recursion (only, if any) has completed
                        # and we are already in the common namespace, thus
                        # we need to pass in the value from the definition site.
                        # Note that since len(nss_call_tails) == 1,
                        # there will not be any call site recursion in this case.
                        var_slash = nss_def_tails[0] + '/' + '___'.join(nss_def_tails[1:] + [var])
                        steps[i]['in'][arg_key] = {'source': var_slash}
                    elif len(nss_call_tails) > 1:
                        inputs_workflow.update({in_name: in_dict})
                        # Store explicit edge call site info up through the recursion.
                        d = {in_name: explicit_edge_defs_copy[arg_val]}
                        # d = {in_name, (namespaces + [step_name_or_key], var)} # ???
                        explicit_edge_calls_copy.update(d)
                        steps[i]['in'][arg_key] = {'source': in_name}
                    else:
                        if nss_def == nss_call:
                            raise Exception("Error! Cannot self-reference the same step!\n" +
                                            f'nss_def {nss_def}\n nss_call {nss_call}')
                        else:
                            # Since nss_call includes step_name_or_key, this should never happen...
                            raise Exception("Error! len(nss_call_tails) == 0! Please file a bug report!\n" +
                                            f'nss_def {nss_def}\n nss_call {nss_call}')

                    arg_keys = [in_name] if in_name in input_mapping_copy else [arg_key]
                    arg_keys = utils.get_input_mappings(input_mapping_copy, arg_keys,
                                                        (arg_key in yaml_tree.get('inputs', {})))

                    out_key_init = '___'.join(nss_def_init + [var])
                    out_key = utils.get_output_mapping(output_mapping_copy, out_key_init)

                    nss_def_embedded = out_key.split('___')[:-1]

                    # NOTE: This if statement is unmotivated and probably masking some other bug, but it works.
                    if out_key.startswith('___'.join(nss_def_init)):
                        nss_def = nss_def_embedded

                    # Add an edge, but in a carefully chosen subgraph.
                    # If you add an edge whose head/tail is outside of the subgraph,
                    # graphviz may segfault! Moreover, even if graphviz doesn't
                    # segfault, adding an edge in a given subgraph can cause the
                    # nodes themselves to be rendered in that subgraph, even
                    # though the nodes are defined in a different subgraph!
                    # The correct thing to do is to use the graph associated with
                    # the lowest_common_ancestor of the definition and call site.
                    # (This is the only reason we need to pass in all subgraphs.)
                    label = var.split('___')[-1]
                    graph_init = subgraphs[len(nss_def_inits)]
                    # Let's use regular blue for explicit edges.
                    # Use constraint=false ?
                    for arg_key_ in arg_keys:
                        # TODO: Double check that we can use the same graph_init for all edges.
                        # Since input_mapping_copy really just factors edges through a single workflow input,
                        # this should hopefully be correct.

                        # NOTE: This if statement is unmotivated and probably masking some other bug, but it works.
                        nss_call_embedded = arg_key_.split('___')[:-1]
                        if arg_key_.startswith('___'.join(namespaces + [step_name_or_key])):
                            nss_call = nss_call_embedded
                        elif arg_key_.startswith(step_name_or_key):
                            nss_call = namespaces + nss_call_embedded
                        else:
                            nss_call = namespaces + [step_name_or_key] + nss_call_embedded

                        utils_graphs.add_graph_edge(args, graph_init, nss_def, nss_call, label, color='blue')
            elif isinstance(arg_val, Dict) and 'wic_inline_input' in arg_val:
                arg_val = arg_val['wic_inline_input']

                if arg_key in steps[i].get('scatter', []):
                    # Promote scattered input types to arrays
                    in_dict['type'] = {'type': 'array', 'items': in_dict['type']}

                inputs_workflow.update({in_name: in_dict})
                in_dict = {**in_dict, 'value': arg_val}
                inputs_file_workflow.update({in_name: in_dict})
                new_val = {'source': in_name}
                steps[i]['in'][arg_key] = new_val

                if args.graph_show_inputs:
                    input_node_name = '___'.join(namespaces + [step_name_or_key, arg_key])
                    attrs = {'label': arg_key, 'shape': 'box', 'style': 'rounded, filled', 'fillcolor': 'lightgreen'}
                    graph_gv.node(input_node_name, **attrs)
                    font_edge_color = 'black' if args.graph_dark_theme else 'white'
                    graph_gv.edge(input_node_name, step_node_name, color=font_edge_color)
                    graph_nx.add_node(input_node_name)
                    graph_nx.add_edge(input_node_name, step_node_name)
                    graphdata.nodes.append((input_node_name, attrs))
                    graphdata.edges.append((input_node_name, step_node_name, {}))
            else:
                arg_var: str = arg_val
                # Leave un-evaluated, i.e. allow the user to inject raw CWL.
                # The un-evaluated string should refer to either an inputs: variable
                # or an internal CWL dependency, i.e. an output from a previous step.
                # Since we do not want to be in the business of parsing raw CWL, if
                # the former is not true, throw an error (with a CLI override).
                # With the change from treating all values as inputs by default,
                # to requiring the !ii inline input custom yaml tag, it is very
                # likely that this code path is simply due to a missing !ii tag.

                # TODO: check this first comment
                # Subworkflows which use workflow inputs: variables cannot
                # (yet) be inlined. Somehow, if they are not marked with
                # inlineable: False, test_inline_subworkflows can still pass.
                # This Exception will (correctly) cause such inlineing tests to fail.
                hashable = False
                inputs_key_dict = {}
                try:
                    hash(arg_var)
                    hashable = True
                    inputs_key_dict = yaml_tree['inputs'][arg_var]
                except Exception:
                    pass

                if not args.allow_raw_cwl and (not hashable or arg_var not in yaml_tree.get('inputs', {})):
                    if not args.allow_raw_cwl:
                        print(f"Warning! Did you forget to use !ii before {arg_var} in {yaml_stem}.wic?")
                        print('If you want to compile the workflow anyway, use --allow_raw_cwl')
                        sys.exit(1)

                    inputs = yaml_tree.get('inputs', {})
                    unbound_lit_var = 'Error! Unbound literal variable'
                    if inputs == {}:
                        raise Exception(f"{unbound_lit_var}{arg_var} not in inputs: tag in {yaml_stem}.wic")
                    inputs_dump = yaml.dump({'inputs': inputs})
                    raise Exception(f"{unbound_lit_var}{arg_var} not in\n{inputs_dump}\nin {yaml_stem}.wic")

                if 'doc' in inputs_key_dict:
                    inputs_key_dict['doc'] += '\\n' + in_dict.get('doc', '')
                else:
                    inputs_key_dict['doc'] = in_dict.get('doc', '')
                if 'label' in inputs_key_dict:
                    inputs_key_dict['label'] += '\\n' + in_dict.get('label', '')
                else:
                    inputs_key_dict['label'] = in_dict.get('label', '')

                if not hashable:
                    pass  # ???
                elif arg_var in input_mapping_copy:
                    input_mapping_copy[arg_var].append(in_name)
                else:
                    input_mapping_copy[arg_var] = [in_name]
                # TODO: We can use un-evaluated variable names for input mapping; no notation for output mapping!
                steps[i]['in'][arg_key] = arg_var  # Leave un-evaluated

        for arg_key in args_required:
            # print('arg_key', arg_key)
            in_name = f'{step_name_or_key}___{arg_key}'
            if arg_key in args_provided:
                continue  # We already covered this case above.
            if in_name in inputs_file_workflow:
                # We provided an explicit argument (but not an edge) in a subworkflow,
                # and now we just need to pass it up to the root workflow.
                # print('passing', in_name)
                in_dict = utils_cwl.copy_cwl_input_output_dict(in_tool[arg_key])
                inputs_workflow.update({in_name: in_dict})
                arg_keyval = {arg_key: in_name}
                steps[i] = utils_cwl.add_yamldict_keyval_in(steps[i], step_key, arg_keyval)

                # Obviously since we supplied a value, we do NOT want to perform edge inference.
                continue
            if arg_key in sub_args_provided:  # Edges have been explicitly provided
                # The definition site recursion (if any) and the call site
                # recursion (yes, see above), have both completed and we are
                # now in the common namespace, thus
                # we need to pass in the value from the definition site.
                # Extract the stored defs namespaces from explicit_edge_calls.
                # (See massive comment above.)
                (nss_def_init, var) = explicit_edge_calls_copy[arg_key]

                nss_def_embedded = var.split('___')[:-1]
                nss_call_embedded = arg_key.split('___')[:-1]
                nss_def = nss_def_init + nss_def_embedded
                # [step_name_or_key] is correct; nss_def_init already contains it from the recursive call
                nss_call = namespaces + [step_name_or_key] + nss_call_embedded

                nss_def_inits, nss_def_tails = utils.partition_by_lowest_common_ancestor(nss_def, nss_call)
                nss_call_inits, nss_call_tails = utils.partition_by_lowest_common_ancestor(nss_call, nss_def)
                assert nss_def_inits == nss_call_inits

                nss_call_tails_stems = [utils.parse_step_name_str(x)[0] for x in nss_call_tails]
                arg_val = steps[i]['in'][arg_key]
                more_recursion = yaml_stem in nss_call_tails_stems and nss_call_tails_stems.index(yaml_stem) > 0
                if (nss_call_tails_stems == []) or more_recursion:
                    # i.e. (if 'dummy' value) or (if it is possible to do more recursion)
                    in_dict = utils_cwl.copy_cwl_input_output_dict(in_tool[arg_key])
                    inputs_workflow.update({in_name: in_dict})
                    steps[i]['in'][arg_key] = {'source': in_name}
                    # Store explicit edge call site info up through the recursion.
                    explicit_edge_calls_copy.update({in_name: explicit_edge_calls_copy[arg_key]})
                else:
                    # TODO: Check this comment.
                    # The definition site recursion (only, if any) has completed
                    # and we are already in the common namespace, thus
                    # we need to pass in the value from the definition site.
                    # Note that since len(nss_call_tails) == 1,
                    # there will not be any call site recursion in this case.
                    var_slash = nss_def_tails[0] + '/' + '___'.join(nss_def_tails[1:] + [var])
                    steps[i]['in'][arg_key] = {'source': var_slash}

                # NOTE: We already added an edge to the appropriate subgraph above.
                # TODO: vars_workflow_output_internal?
            else:
                if args.inference_disable:
                    continue
                insertions: List[StepId] = []
                in_name_in_inputs_file_workflow: bool = (in_name in inputs_file_workflow)
                arg_key_in_yaml_tree_inputs: bool = (arg_key in yaml_tree.get('inputs', {}))
                steps[i] = inference.perform_edge_inference(args, tools, tools_lst, steps_keys,
                                                            yaml_stem, i, steps, arg_key, graph, is_root, namespaces,
                                                            vars_workflow_output_internal, input_mapping_copy, output_mapping_copy, inputs_workflow, in_name,
                                                            in_name_in_inputs_file_workflow, arg_key_in_yaml_tree_inputs, insertions, wic_steps, testing)
                # NOTE: For now, perform_edge_inference mutably appends to
                # inputs_workflow and vars_workflow_output_internal.

                # Automatically insert steps
                insertions = list(set(insertions))  # Remove duplicates
                if len(insertions) != 0 and args.insert_steps_automatically:
                    insertion = insertions[0]
                    print('Automaticaly inserting step', insertion, i)
                    if len(insertions) != 1:
                        print('Warning! More than one step! Choosing', insertion)

                    yaml_tree_mod = insert_step_into_workflow(yaml_tree_orig, insertion, tools, i)

                    node_data = NodeData(namespaces, yaml_stem, yaml_tree_mod, yaml_tree, tool_i, {},
                                         explicit_edge_defs_copy2, explicit_edge_calls_copy2,
                                         graph, inputs_workflow, '')
                    rose_tree = RoseTree(node_data, rose_tree_list)
                    env_data = EnvData(input_mapping_copy, output_mapping_copy,
                                       inputs_file_workflow, vars_workflow_output_internal,
                                       explicit_edge_defs_copy, explicit_edge_calls_copy)
                    compiler_info = CompilerInfo(rose_tree, env_data)
                    # node_data_dummy = NodeData(None, None, yaml_tree_mod, None, None, None, None, None, None, None)
                    # compiler_info_dummy = CompilerInfo(RoseTree(node_data_dummy, None), None)
                    return compiler_info

        # Add CommandLineTool/Subworkflow outputs tags to workflow out tags.
        # Note: Add all output tags for now, but depending on config options,
        # not all output files will be generated. This may cause an error.
        out_keyvals = {}
        for out_key, out_dict in tool_i.cwl['outputs'].items():
            out_keyvals[out_key] = utils_cwl.copy_cwl_input_output_dict(out_dict)
            # print(out_key, out_keyvals[out_key])
        if not out_keyvals:  # FYI out_keyvals should never be {}
            print(f'Error! no outputs for step {step_key}')
        outputs_workflow.append(out_keyvals)

        steps[i] = utils_cwl.add_yamldict_keyval_out(steps[i], step_key, list(tool_i.cwl['outputs'].keys()))

        if args.partial_failure_enable:
            when_null_clauses = []
            for arg_in in args_required:
                when_null_clauses.append(f'inputs["{arg_in}"] != null')
            when_clause = ' && '.join(when_null_clauses)
            if when_null_clauses:
                if 'when' in steps[i]:
                    print('Warning! overwriting an existing "when" clause')
                steps[i]['when'] = f"$({when_clause})"
        # print()

    # NOTE: add_subgraphs currently mutates graph
    wic_graphviz = wic['wic'].get('graphviz', {})
    ranksame_strs = wic_graphviz.get('ranksame', [])
    ranksame_pairs = [utils.parse_int_string_tuple(x) for x in ranksame_strs]
    steps_ranksame = []
    for num, name in ranksame_pairs:
        step_name_num = utils.step_name_str(yaml_stem, num-1, name)
        step_name_nss = '___'.join(namespaces + [step_name_num])
        steps_ranksame.append(f'"{step_name_nss}"')  # Escape with double quotes.
    utils_graphs.add_subgraphs(args, graph, sibling_subgraphs, namespaces, step_1_names, steps_ranksame)
    step_name_1 = utils.get_step_name_1(step_1_names, yaml_stem, namespaces, steps_keys, subkeys)

    # Add the provided workflow inputs to the workflow inputs from each step
    inputs_combined = {**yaml_tree.get('inputs', {}), **inputs_workflow}
    yaml_tree.update({'inputs': inputs_combined})

    # NOTE: This is a nasty hack because we don't have any syntax for mapping workflow outputs.
    for k, v in yaml_tree.get('outputs', {}).items():
        # Assume the user has manually added the correct namespaced CWL dependency.
        output_mapping_copy[k] = v['outputSource'].replace('/', '___')

    vars_workflow_output_internal = list(set(vars_workflow_output_internal))  # Get uniques
    # (Why are we getting uniques?)
    workflow_outputs = utils_cwl.get_workflow_outputs(args, namespaces, is_root, yaml_stem,
                                                      steps, outputs_workflow, vars_workflow_output_internal,
                                                      graph, tools_lst, step_node_name, tools)
    # Add the provided workflow outputs to the workflow outputs from each step
    outputs_combined = {**yaml_tree.get('outputs', {}), **workflow_outputs}
    yaml_tree.update({'outputs': outputs_combined})

    # NOTE: currently mutates yaml_tree (maybe)
    utils_cwl.maybe_add_requirements(yaml_tree, steps_keys, wic_steps, subkeys)

    # Finally, rename the steps to be unique
    # and convert the list of steps into a dict / list
    steps_dict = {}
    steps_list = []
    for i, step_key in enumerate(steps_keys):
        wic_step_i = wic_steps.get(f'({i+1}, {step_key})', {})
        plugin_ns_i = wic_step_i.get('wic', {}).get('namespace', 'global')
        stepid = StepId(Path(step_key).stem, plugin_ns_i)

        step_name_i = utils.step_name_str(yaml_stem, i, step_key)
        step_name_or_key = step_name_i if step_key.endswith('.wic') or stepid in tools else step_key
        # steps[i] = {steps[i], 'id': step_name_or_key}
        step_i_copy = {**steps[i]}
        del step_i_copy['id']
        steps_dict.update({step_name_or_key: step_i_copy})
        step_i_copy = {'id': step_name_or_key, **step_i_copy}
        steps_list.append(step_i_copy)
    yaml_tree.update({'steps': steps_list})  # steps_list ?

    # Dump the workflow inputs to a separate yml file.
    yaml_inputs: WorkflowInputsFile = {}
    for key, in_dict in inputs_file_workflow.items():
        new_keyval: WorkflowInputsFile = {}
        if 'File' == in_dict['type']:
            # path = Path(in_dict['value']).name # NOTE: Use .name ?
            newval = {'class': 'File', 'path': in_dict['value']}
            if 'format' in in_dict:
                in_format = in_dict['format']
                if isinstance(in_format, List):
                    in_format = list(set(in_format))  # get uniques
                    if len(in_format) > 1:
                        print(f'NOTE: More than one input file format for {key}')
                        print(f'formats: {in_format}')
                        print(f'Choosing {in_format[0]}')
                    in_format = in_format[0]
                newval['format'] = in_format
            new_keyval = {key: newval}
        elif 'Directory' == in_dict['type']:
            dir = Path(in_dict['value'])
            if not dir.is_absolute():
                dir = Path('autogenerated') / dir
            dir.mkdir(parents=True, exist_ok=True)
            newval = {'class': 'Directory', 'location': in_dict['value']}
            new_keyval = {key: newval}
        # TODO: Check for all valid types?
        else:
            # We cannot store string values as a dict, so use type: ignore
            arg_val = in_dict['value']
            new_keyval = {key: arg_val}
        # else:
        #    raise Exception(f"Error! Unknown type: {in_dict['type']}")
        yaml_inputs.update(new_keyval)

    if not testing:
        print('finishing', ('  ' * len(namespaces)) + yaml_path)
    # Note: We do not necessarily need to return inputs_workflow.
    # 'Internal' inputs are encoded in yaml_tree. See Comment above.
    node_data = NodeData(namespaces, yaml_stem, yaml_tree_orig, yaml_tree, tool_i, yaml_inputs,
                         explicit_edge_defs_copy2, explicit_edge_calls_copy2,
                         graph, inputs_workflow, step_name_1)
    rose_tree = RoseTree(node_data, rose_tree_list)
    env_data = EnvData(input_mapping_copy, output_mapping_copy, inputs_file_workflow, vars_workflow_output_internal,
                       explicit_edge_defs_copy, explicit_edge_calls_copy)
    compiler_info = CompilerInfo(rose_tree, env_data)
    return compiler_info


def insert_step_into_workflow(yaml_tree_orig: Yaml, stepid: StepId, tools: Tools, i: int) -> Yaml:
    """Inserts the step with given stepid into a workflow at the given index.

    Args:
        yaml_tree_orig (Yaml): The original Yaml tree
        stepid (StepId): The name of the workflow step to be inserted.
        tools (Tools): The CWL CommandLineTool definitions found using get_tools_cwl().\n
        yml files that have been compiled to CWL SubWorkflows are also added during compilation.
        i (int): The index to insert the new workflow step

    Returns:
        Yaml: A modified Yaml tree with the given stepid inserted at index i
    """
    yaml_tree_mod = yaml_tree_orig
    steps_mod: List[Yaml] = yaml_tree_mod['steps']
    steps_mod.insert(i, {stepid.stem: None})

    # Add inference rules annotations (i.e. for insertions)
    tool = tools[stepid]
    out_tool = tool.cwl['outputs']

    inference_rules_dict = {}
    for out_key, out_val in out_tool.items():
        if 'format' in out_val:
            inference_rules_dict[out_key] = inference_rules.get(out_val['format'], 'default')
    inf_dict = {'wic': {'inference': inference_rules_dict}}
    keystr = f'({i+1}, {stepid.stem})'  # The yml file uses 1-based indexing

    if 'wic' in yaml_tree_mod:
        if 'steps' in yaml_tree_mod['wic']:
            yaml_tree_mod['wic']['steps'] = utils.reindex_wic_steps(yaml_tree_mod['wic']['steps'], i+1)
            yaml_tree_mod['wic']['steps'][keystr] = inf_dict
        else:
            yaml_tree_mod['wic'].update({'steps': {keystr: inf_dict}})
    else:
        yaml_tree_mod.update({'wic': {'steps': {keystr: inf_dict}}})
    return yaml_tree_mod



from typing import Any, Dict, Type

import yaml

# NOTE: In the following constructors, you CANNOT return the exact same yaml tag
# Otherwise, the loader is not idempotent. Specifically, then other tooling
# (i.e. the python api) cannot simply emit the dictionaries returned here,
# because then these constructors will fire again.


def anchor_constructor(loader: yaml.SafeLoader, node: yaml.nodes.ScalarNode) -> Dict[str, Any]:
    val = loader.construct_scalar(node)
    name = 'wic_anchor'  # NOT '!&'
    return {name: val}


def alias_constructor(loader: yaml.SafeLoader, node: yaml.nodes.ScalarNode) -> Dict[str, Any]:
    val = loader.construct_scalar(node)
    name = 'wic_alias'  # NOT '!*'
    return {name: val}


def inlineinput_constructor(loader: yaml.SafeLoader, node: yaml.nodes.Node) -> Dict[str, Dict[str, Any]]:
    val: Any
    if isinstance(node, yaml.nodes.ScalarNode):
        try:
            # loader.construct_scalar always returns a string, whereas
            val = yaml.safe_load(node.value)
            # yaml.safe_load returns the correct primitive types
        except Exception:
            # but fallback to a string if it is not actually a primitive type.
            val = loader.construct_scalar(node)
    elif isinstance(node, yaml.nodes.MappingNode):
        val = loader.construct_mapping(node)
    elif isinstance(node, yaml.nodes.SequenceNode):
        val = loader.construct_sequence(node)
    else:
        raise Exception(f'Unknown yaml node type! {node}')
    name = 'wic_inline_input'  # NOT '!ii'
    return {name: val}


def wic_loader() -> Type[yaml.SafeLoader]:
    loader = yaml.SafeLoader
    loader.add_constructor("!&", anchor_constructor)
    loader.add_constructor("!*", alias_constructor)
    loader.add_constructor("!ii", inlineinput_constructor)
    return loader



from pathlib import Path
import time
from typing import Dict

from hypothesis.strategies import SearchStrategy
import hypothesis_jsonschema as hj

import sophios
import sophios.cli
import sophios.input_output as io
import sophios.plugins
import sophios.schemas
import sophios.schemas.wic_schema
import sophios.utils
import sophios.api
import sophios.api.pythonapi
from sophios.wic_types import Json, Yaml


args = sophios.cli.get_args()
# Just read from the disk and pass around config object
global_config = io.get_config(Path(args.config_file), Path(args.config_file))
tools_cwl = sophios.plugins.get_tools_cwl(global_config, quiet=args.quiet)
sophios.api.pythonapi.global_config = tools_cwl  # Use path fallback in the CI
yml_paths = sophios.plugins.get_yml_paths(global_config)


yaml_stems = sophios.utils.flatten([list(p) for p in yml_paths.values()])
schema_store: Dict[str, Json] = {}
validator = sophios.schemas.wic_schema.get_validator(tools_cwl, yaml_stems, schema_store, write_to_disk=True)

yml_paths_tuples = [(yml_path_str, yml_path)
                    for yml_namespace, yml_paths_dict in yml_paths.items()
                    for yml_path_str, yml_path in yml_paths_dict.items()]


for yml_path_str, yml_path in yml_paths_tuples:
    schema = sophios.schemas.wic_schema.compile_workflow_generate_schema(args.homedir, yml_path_str, yml_path,
                                                                         tools_cwl, yml_paths, validator,
                                                                         args.ignore_validation_errors,
                                                                         args.allow_raw_cwl)
    # overwrite placeholders in schema_store. See comment in get_validator()
    schema_store[schema['$id']] = schema

validator = sophios.schemas.wic_schema.get_validator(tools_cwl, yaml_stems, schema_store, write_to_disk=True)


def wic_yaml_filter_blank_steps(yml: Yaml) -> bool:
    """Filters out Yaml instances with blank steps.

    Args:
        yml (Yaml): A randomly generated Yaml instance.

    Returns:
        bool: True if there are no blank steps.
    """
    steps = yml.get('steps', [])
    return not (steps == [] or any([step == {} for step in steps]))


def wic_yaml_filter_implementations_or_steps(yml: Yaml) -> bool:
    """Filters out Yaml instances with no implementations and no steps.

    Args:
        yml (Yaml): A randomly generated Yaml instance.

    Returns:
        bool: True if there is either implementations or steps (or both).
    """
    return ('implementations' in yml or 'steps' in yml)


time_initial = time.time()

wic_schema = sophios.schemas.wic_schema.wic_main_schema(tools_cwl, yaml_stems, schema_store, hypothesis=True)
wic_strategy: SearchStrategy = hj.from_schema(wic_schema)
# NOTE: The CLI version of mypy and the VSCode version of mypy disagree on the
# following line. The "type: ignore" comment is NOT unused.
wic_strategy = wic_strategy.filter(wic_yaml_filter_blank_steps)
wic_strategy = wic_strategy.filter(wic_yaml_filter_implementations_or_steps)

time_final = time.time()
print(f'from_schema time: {round(time_final - time_initial, 4)} seconds')
print()



# pylint: disable=W1203
"""CLT utilities."""
import logging
from pathlib import Path
import subprocess as sub
from typing import Any, ClassVar, Optional, TypeVar, Union

import cwl_utils.parser as cu_parser
import yaml
from mergedeep import merge, Strategy
from cwl_utils.parser import CommandLineTool as CWLCommandLineTool
from cwl_utils.parser import load_document_by_uri, load_document_by_yaml
from pydantic import BaseModel, ConfigDict, Field, PrivateAttr, field_validator

from sophios import compiler, input_output, plugins, utils_cwl
from sophios import run_local as run_local_module
from sophios.cli import get_args
from sophios.utils_graphs import get_graph_reps
from sophios.wic_types import CompilerInfo, RoseTree, StepId, Tool, Tools, YamlTree

from ._types import ScatterMethod


global_config: Tools = {}


logger = logging.getLogger("WIC Python API")


class DisableEverythingFilter(logging.Filter):
    # pylint:disable=too-few-public-methods
    def filter(self, record: logging.LogRecord) -> bool:
        return False


# Based on user feedback,
# disable any and all warnings coming from autodiscovery.
logger_wicad = logging.getLogger("wicautodiscovery")
logger_wicad.addFilter(DisableEverythingFilter())


_WIC_PATH = Path(__file__).parent.parent.parent.parent  # WIC dir


class InvalidInputValueError(Exception):
    pass


class MissingRequiredValueError(Exception):
    pass


class InvalidStepError(Exception):
    pass


class InvalidLinkError(Exception):
    pass


class InvalidCLTError(ValueError):
    pass


CWLInputParameter = Union[
    cu_parser.cwl_v1_0.CommandInputParameter,
    cu_parser.cwl_v1_1.CommandInputParameter,
    cu_parser.cwl_v1_2.CommandInputParameter,
]  # cwl_utils does not have CommandInputParameter union

CWLOutputParameter = Union[
    cu_parser.cwl_v1_0.CommandOutputParameter,
    cu_parser.cwl_v1_1.CommandOutputParameter,
    cu_parser.cwl_v1_2.CommandOutputParameter,
]  # cwl_utils does not have CommandInputParameter union

StrPath = TypeVar("StrPath", str, Path)


class ProcessInput(BaseModel):  # pylint: disable=too-few-public-methods
    """Input of CWL CommandLineTool or Workflow."""

    inp_type: Any
    name: str
    # NOTE: Not optional, but we can't initialize it yet
    parent_obj: Any = Field(default=None)  # Process: Union[Step, Workflow]
    value: Any = Field(default=None)  # validation happens at assignment
    required: bool = True
    linked: bool = False

    def __init__(self, name: str, inp_type: Any) -> None:
        inp_type = utils_cwl.canonicalize_type(inp_type)
        if isinstance(inp_type, list) and "null" in inp_type:
            required = False
        else:
            required = True
        super().__init__(inp_type=inp_type, name=name, required=required)

    @field_validator("name", mode="before")
    @classmethod
    def get_name_from_id(cls, cwl_id: str) -> Any:
        """Return name of input from InputParameter.id."""
        return cwl_id.split("#")[-1]

    @field_validator("inp_type", mode="before")
    @classmethod
    def set_inp_type(cls, inp: Any) -> Any:
        """Return inp_type."""
        if isinstance(inp, list):  # optional inps
            inp = inp[1]
        return inp

    def _set_value(
        self, __value: Any, linked: bool = False
    ) -> None:
        """Set input value."""
        self.value = __value  # to be used for linking inputs */&
        if linked:
            self.linked = True


class ProcessOutput(BaseModel):  # pylint: disable=too-few-public-methods
    """Output of CWL CommandLineTool or Workflow."""

    out_type: Any
    name: str
    # NOTE: Not optional, but we can't initialize it yet
    parent_obj: Any = Field(default=None)  # Process: Union[Step, Workflow]
    value: Any = Field(default=None)  # validation happens at assignment
    required: bool = True
    linked: bool = False

    def __init__(self, name: str, out_type: Any) -> None:
        out_type = utils_cwl.canonicalize_type(out_type)
        if isinstance(out_type, list) and "null" in out_type:
            required = False
        else:
            required = True
        super().__init__(out_type=out_type, name=name, required=required)

    @field_validator("name", mode="before")
    @classmethod
    def get_name_from_id(cls, cwl_id: str) -> Any:
        """Return name of input from InputParameter.id."""
        return cwl_id.split("#")[-1]

    @field_validator("out_type", mode="before")
    @classmethod
    def set_out_type(cls, out: Any) -> Any:
        """Return out_type."""
        if isinstance(out, list):  # optional outs
            out = out[1]
        return out

    def _set_value(
        self, __value: Any, linked: bool = False
    ) -> None:
        """Set output value."""
        self.value = __value  # to be used for linking inputs */&
        if linked:
            self.linked = True


def _default_dict() -> dict:
    return {}


def _get_value_from_cfg(value: Any) -> Any:  # validation happens in Step
    if isinstance(value, dict):
        if "Directory" in value.values():
            try:
                value_ = Path(value["location"])
            except BaseException as exc:
                raise InvalidInputValueError() from exc
            if not value_.is_dir():
                raise InvalidInputValueError(f"{str(value_)} is not a directory")
            return value_
        return value
    return value


# Process = Union[Step, Workflow]


def set_input_Step_Workflow(process_self: Any, __name: str, __value: Any) -> Any:
    index = process_self._input_names.index(__name)
    if isinstance(__value, ProcessOutput):
        process_other = __value.parent_obj
        if not __value.linked:
            try:
                local_input = process_self.inputs[index]
                # NOTE: Relax exact equality for Any type
                # if not local_input.inp_type == __value.out_type and not local_input.inp_type == Any and not __value.out_type == Any:
                #     raise InvalidLinkError(
                #         f"links must have the same input type. "
                #         f"cannot link {local_input.name} to {__value.name}"
                #         f"with types {local_input.inp_type} to {__value.out_type}"
                #     )
                if isinstance(process_other, Workflow):
                    tmp = __value.name  # Use the formal parameter / variable name
                    local_input._set_value(f"{tmp}", linked=True)
                    __value._set_value(f"{tmp}", linked=True)
                else:
                    # Use the current value so we can exactly reproduce hand-crafted wic files.
                    # (Very useful for regression testing!)
                    # NOTE: process_name is either clt name or workflow name
                    tmp = __value.value if __value.value else f"{__name}{process_self.process_name}"
                    alias_dict = {'wic_alias': tmp}
                    local_input._set_value(alias_dict, linked=True)
                    anchor_dict = {'wic_anchor': tmp}
                    __value._set_value(anchor_dict, linked=True)
            except BaseException as exc:
                raise exc
        else:  # value is already linked to another inp
            try:
                local_input = process_self.inputs[index]
                # NOTE: Relax exact equality for Any type
                # if not local_input.inp_type == __value.out_type and not local_input.inp_type == Any and not __value.out_type == Any:
                #     raise InvalidLinkError(
                #         f"links must have the same input type. "
                #         f"cannot link {local_input.name} to {__value.name} "
                #         f"with types {local_input.inp_type} to {__value.out_type}"
                #     )
                if isinstance(process_other, Workflow):
                    tmp = __value.name  # Use the formal parameter / variable name
                    local_input._set_value(f"{tmp}", linked=True)
                    __value._set_value(f"{tmp}", linked=True)
                else:
                    anchor_dict = __value.value
                    alias_dict = {'wic_alias': anchor_dict['wic_anchor']}
                    local_input._set_value(alias_dict, linked=True)
            except BaseException as exc:
                raise exc

    else:
        # obj = process_self.inputs[index]
        # NOTE: "TypeError: typing.Any cannot be used with isinstance()"
        # if not obj.inp_type == Any and not isinstance(__value, obj.inp_type) and not isinstance(__value, list):
        #     raise TypeError(
        #         f"invalid attribute type for {obj.name}: "
        #         f"got {__value.__class__.__name__}, "
        #         f"expected {obj.inp_type.__name__}"
        #     )
        ii_dict = {'wic_inline_input': __value}
        process_self.inputs[index]._set_value(ii_dict)


class Step(BaseModel):  # pylint: disable=too-few-public-methods
    """Base class for Step of Workflow."""
    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)

    clt: CWLCommandLineTool
    clt_path: Path
    process_name: str
    cwl_version: str
    inputs: list[ProcessInput]
    outputs: list[ProcessOutput]
    yaml: dict[str, Any]
    cfg_yaml: dict = Field(default_factory=_default_dict)

    # these are not part of 'clt data'
    scatter: list[ProcessInput] = []
    scatterMethod: str = ''
    # use when tag to enable conditional steps
    when: str = ''
    _input_names: list[str] = PrivateAttr(default_factory=list)
    _output_names: list[str] = PrivateAttr(default_factory=list)

    def __init__(self, clt_path: StrPath, config_path: Optional[StrPath] = None):
        # validate using cwl.utils
        if not isinstance(clt_path, (Path, str)):
            raise TypeError("cwl_path must be a Path or str")
        clt_path_ = Path(clt_path) if isinstance(clt_path, str) else clt_path
        stepid = StepId(clt_path_.stem, 'global')

        if clt_path_.exists():
            try:
                clt = load_document_by_uri(clt_path_)
            except Exception as exc:
                raise InvalidCLTError(f"invalid cwl file: {clt_path_}") from exc
            with clt_path_.open("r", encoding="utf-8") as file:
                yaml_file = yaml.safe_load(file)

            # Add the Tool to the global config dictionary. See explanation below.
            global_config[stepid] = Tool(clt_path_.stem, yaml_file)
        elif stepid in global_config:
            # Use the path fallback mechanism.
            # In other words, if the hardcoded, nonportable, system-dependent paths
            # are not found above (i.e. because someone else is trying to run
            # the workflow on another machine), then try to load the CLT from
            # a global dictionary, which can be pre-configured to have paths
            # that exist on the target machine.
            # Again, global_config is just a dictionary; it couldn't be more pythonic.
            # You can initialize it however you want (or not at all!).
            tool = global_config[stepid]

            logger.info(f'{clt_path_} does not exist, but {clt_path_.stem} was found in the global config.')
            logger.info(f'Using file contents from {tool.run_path}')

            yaml_file = tool.cwl
            clt = load_document_by_yaml(yaml_file, tool.run_path)
        else:
            logger.warning(f'Warning! {clt_path_} does not exist, and')
            logger.warning(f'{clt_path_.stem} was not found in the global config.')
            raise InvalidCLTError(f"invalid cwl file: {clt_path_}")

        if config_path:
            cfg_path_ = Path(config_path) if isinstance(config_path, str) else config_path
            with cfg_path_.open("r", encoding="utf-8") as file:
                cfg_yaml = yaml.safe_load(file)
        else:
            cfg_yaml = _default_dict()  # redundant, to avoid it being unbound
        process_name = clt_path_.stem
        data = {
            "clt": clt,
            "clt_path": clt_path_,
            "cwl_version": clt.cwlVersion,
            "process_name": process_name,
            "inputs": clt.inputs,
            "outputs": clt.outputs,
            "yaml": yaml_file,
            "cfg_yaml": cfg_yaml,
        }
        super().__init__(**data)
        for inp in self.inputs:
            inp.parent_obj = self  # Create a reference to the parent Step object
        self._input_names = [inp.id.split("#")[-1] for inp in clt.inputs]
        for out in self.outputs:
            out.parent_obj = self  # Create a reference to the parent Step object
        self._output_names = [out.id.split("#")[-1] for out in clt.outputs]
        if config_path:
            self._set_from_io_cfg()

    @field_validator("inputs", mode="before")
    @classmethod
    def cast_to_process_input_model(
        cls, cwl_inps: list[CWLInputParameter]
    ) -> list[ProcessInput]:
        """Populate inputs from cwl.inputs."""
        # NOTE: .type in version 0.31 only!
        # NOTE: Cannot initialize .parent_obj = self here due to @classmethod
        return [ProcessInput(str(x.id), x.type_) for x in cwl_inps]

    @field_validator("outputs", mode="before")
    @classmethod
    def cast_to_process_output_model(
        cls, cwl_outs: list[CWLOutputParameter]
    ) -> list[ProcessOutput]:
        """Populate outputs from cwl.outputs."""
        # NOTE: .type in version 0.31 only!
        # NOTE: Cannot initialize .parent_obj = self here due to @classmethod
        return [ProcessOutput(str(x.id), x.type_) for x in cwl_outs]

    def __repr__(self) -> str:
        repr_ = f"Step(clt_path={self.clt_path.__repr__()})"
        return repr_

    def __setattr__(self, __name: str, __value: Any) -> Any:
        if __name in ["inputs", "outputs", "yaml", "cfg_yaml", "process_name", "_input_names", "_output_names",
                      "__private_attributes__", "__pydantic_private__"]:
            return super().__setattr__(__name, __value)
        if __name == "scatterMethod":
            if hasattr(ScatterMethod, __value):
                return super().__setattr__(__name, __value)
            else:
                raise ValueError(
                    f"Invalid value for scatterMethod the valid values are : \n {ScatterMethod.dotproduct.value} "
                    f"{ScatterMethod.flat_crossproduct.value} {ScatterMethod.nested_crossproduct.value}\n")
        if __name == "scatter":
            if not all([isinstance(x, ProcessInput) for x in __value]):
                raise TypeError("all scatter inputs must be ProcessInput type")
            return super().__setattr__(__name, __value)
        if __name == "when":
            if (isinstance(__value, str)) and __value.startswith('$(') and __value.endswith(')'):
                return super().__setattr__(__name, __value)
            else:
                raise ValueError("Invalid input to when.The js string must start with '$(' and end with ')'")

        if hasattr(self, "_input_names") and __name in self._input_names:
            set_input_Step_Workflow(self, __name, __value)
        else:
            return super().__setattr__(__name, __value)

    def __getattr__(self, __name: str) -> Any:
        if __name in ["__pydantic_private__", "__class__", "__private_attributes__"]:
            return super().__getattribute__(__name)
        if __name != "_output_names" and __name in self._output_names:  # and hasattr(self, "_output_names") ?
            return self.outputs[self._output_names.index(__name)]
        # https://github.com/pydantic/pydantic/blob/812516d71a8696d5e29c5bdab40336d82ccde412/pydantic/main.py#L743-744
        # "We put `__getattr__` in a non-TYPE_CHECKING block because otherwise, mypy allows arbitrary attribute access
        # The same goes for __setattr__ and __delattr__, see: https://github.com/pydantic/pydantic/issues/8643"
        return super().__getattr__(__name)  # type: ignore

    def _set_from_io_cfg(self) -> None:
        for name, value in self.cfg_yaml.items():
            value_ = _get_value_from_cfg(value)
            setattr(self, name, value_)

    def _validate(self) -> None:
        for inp in self.inputs:
            if inp.required and inp.value is None:
                raise MissingRequiredValueError(f"{inp.name} is required")

    @property
    def _yml(self) -> dict:
        in_dict: dict[str, Any] = {}  # NOTE: input values can be arbitrary JSON; not just strings!
        for inp in self.inputs:
            if inp.value is not None:
                if isinstance(inp.value, Path):
                    # Special case for Path since it does not inherit from YAMLObject
                    in_dict[inp.name] = str(inp.value)
                elif isinstance(inp.value, dict) and isinstance(inp.value.get('wic_alias', {}), Path):
                    # Special case for Path since it does not inherit from YAMLObject
                    in_dict[inp.name] = {'wic_alias': str(inp.value['wic_alias'])}
                elif isinstance(inp.value, dict) and isinstance(inp.value.get('wic_inline_input', {}), Path):
                    # Special case for Path since it does not inherit from YAMLObject
                    in_dict[inp.name] = {'wic_inline_input': str(inp.value['wic_inline_input'])}
                elif isinstance(inp.value, dict) and isinstance(inp.value.get('wic_inline_input', {}), str):
                    # Special case for inline str since it does not inherit from YAMLObject
                    in_dict[inp.name] = {'wic_inline_input': inp.value.get('wic_inline_input')}
                elif isinstance(inp.value, str):
                    in_dict[inp.name] = inp.value  # Obviously strings are serializable
                elif isinstance(inp.value, yaml.YAMLObject):
                    # Serialization and deserialization logic should always be
                    # encapsulated within each object. For the pyyaml library,
                    # each object should inherit from pyyaml.YAMLObject.
                    # See https://pyyaml.org/wiki/PyYAMLDocumentation
                    # Section "Constructors, representers, resolvers"
                    # class Monster(yaml.YAMLObject): ...
                    in_dict[inp.name] = inp.value
                else:
                    logger.warning(f'Warning! input name {inp.name} input value {inp.value}')
                    logger.warning('is not an instance of YAMLObject. The default str() serialization')
                    logger.warning('logic often gives bad results. Please explicitly inherit from YAMLObject.')
                    in_dict[inp.name] = inp.value

        out_list: list = []  # The out: tag is a list, not a dict
        out_list = [{out.name: out.value} for out in self.outputs if out.value]
        # list of inputs to be scattered on
        scatter_list: list = []
        scatter_list = [sc_inp.name for sc_inp in self.scatter]

        d = {
            "id": self.process_name,
            "in": in_dict,
            "out": out_list,
        }
        # scatter operates on input sink
        if self.scatter:
            d["scatter"] = scatter_list
            if '' == self.scatterMethod:
                self.scatterMethod = ScatterMethod.dotproduct.value
            d["scatterMethod"] = self.scatterMethod
        # when operates on step
        if self.when != '':
            d["when"] = self.when
        return d

    def _save_cwl(self, path: Path) -> None:
        cwl_adapters = path.joinpath("cwl_adapters")
        cwl_adapters.mkdir(exist_ok=True, parents=True)
        with open(
            cwl_adapters.joinpath(f"{self.process_name}.cwl"),
            "w",
            encoding="utf-8",
        ) as file:
            file.write(yaml.dump(self.yaml))


def extract_tools_paths_NONPORTABLE(steps: list[Step]) -> Tools:
    """extract a Tools global configuration object from the NONPORTABLE paths hardcoded in the Steps.

    Args:
        steps (list[Step]): A list of Steps.

    Returns:
        Tools: A Tools object, which is just a dictionary mapping a globally unique
        (i.e. portable) identifier to a system-dependendent Path.
    """
    return {StepId(step.process_name, 'global'): Tool(str(step.clt_path), step.yaml) for step in steps}


# Analogous to cu_parser.Process, we want to create our own Union. However,
# Process = Union[Step, Workflow]  # Cannot define Process before Workflow


class Workflow(BaseModel):
    model_config: ClassVar[ConfigDict] = ConfigDict(arbitrary_types_allowed=True)

    steps: list  # list[Process]  # and cannot use Process defined after Workflow within a Workflow
    process_name: str
    inputs: list[ProcessInput] = []
    outputs: list[ProcessOutput] = []
    _input_names: list[str] = PrivateAttr(default_factory=list)
    _output_names: list[str] = PrivateAttr(default_factory=list)
    yml_path: Optional[Path] = Field(default=None)
    # Cannot use field() from dataclasses. Otherwise:
    # from dataclasses import field
    # field(default=None, init=False, repr=False)
    # TypeError: 'ModelPrivateAttr' object is not iterable

    def __init__(self, steps: list, workflow_name: str):
        data = {
            "process_name": workflow_name,
            "steps": steps,
        }
        super().__init__(**data)

    def __post_init__(self) -> None:
        self._validate()

    def _validate(self) -> None:
        # Only the root workflow should have all required inputs and can be validated.
        # Cannot validate subworkflows because by definition,
        # subworkflows will NOT have all required inputs.
        for s in self.steps:
            try:
                if isinstance(s, Step):
                    s._validate()  # pylint: disable=W0212
                if isinstance(s, Workflow):
                    # recursively validate subworkflows ?
                    s._validate()  # pylint: disable=W0212
            except BaseException as exc:
                raise InvalidStepError(
                    f"{s.process_name} is missing required inputs"
                ) from exc

    def append(self, step_: Step) -> None:
        """Append step to Workflow."""
        if not isinstance(step_, (Step, Workflow)):
            raise TypeError("step must be either a Step or a Workflow")
        self.steps.append(step_)

        # for name in step_._input_names:
        #     self._input_names.append(name)
        # for inp in step_.inputs:
        #     self.inputs.append(inp)

        # for name in step_._output_names:
        #     self._output_names.append(name)
        # for out in step_.outputs:
        #     self.outputs.append(out)

    @property
    def yaml(self) -> dict[str, Any]:
        """Converts a Workflow to the equivalent underlying WIC YML representation, in-memory."""
        # NOTE: Use the CWL v1.2 `Any` type for lack of actual type information.
        # TODO: try inp.inp_type (if initialized?)
        inputs: dict[str, Any] = {inp.name: {'type': 'Any'} for inp in self.inputs}
        # TODO: outputs?
        steps = []
        for s in self.steps:
            if isinstance(s, Step):
                steps.append(s._yml)
            elif isinstance(s, Workflow):
                ins = {
                    inp.name: inp.value
                    for inp in s.inputs
                    if inp.value is not None  # Subworkflow args are not required
                }
                parentargs: dict[str, Any] = {"in": ins} if ins else {}
                # See the second to last line of ast.read_ast_from_disk()
                d = {'id': self.process_name + '.wic',
                     'subtree': s.yaml,  # recursively call .yaml (i.e. on s, not self)
                     'parentargs': parentargs}
                steps.append(d)
            #  else: ...
        yaml_contents = {"inputs": inputs, "steps": steps} if inputs else {"steps": steps}
        return yaml_contents

    def write_ast_to_disk(self, directory: Path) -> None:
        """Converts a Workflow to the equivalent underlying WIC YML representation,
        and writes each subworkflow to disk as a separate yml file.

        Args:
            directory (Path): The directory to write the yml file(s).
        """
        # NOTE: Use the CWL v1.2 `Any` type for lack of actual type information.
        # TODO: try inp.inp_type (if initialized?)
        inputs: dict[str, Any] = {inp.name: {'type': 'Any'} for inp in self.inputs}
        # TODO: outputs?
        steps = []
        for s in self.steps:
            if isinstance(s, Step):
                steps.append(s._yml)
            elif isinstance(s, Workflow):
                ins = {
                    inp.name: inp.value
                    for inp in s.inputs
                    if inp.value is not None  # Subworkflow args are not required
                }
                parentargs: dict[str, Any] = {"in": ins} if ins else {}
                s.write_ast_to_disk(directory)  # recursively call
                steps.append({'id': s.process_name + '.wic', **parentargs})
            #  else: ...
        yaml_contents = {"inputs": inputs, "steps": steps} if inputs else {"steps": steps}
        # NOTE: For various reasons, process_name should be globally unique.
        # In this case, it is to avoid overwriting files.
        directory.mkdir(exist_ok=True, parents=True)
        with open(str(directory / self.process_name) + '.wic', mode='w', encoding='utf-8') as f:
            f.write(yaml.dump(yaml_contents, sort_keys=False, line_break='\n', indent=2))

    def add_input(self, __name: str) -> Any:
        # Assume the user wants to auto-generate an input on the fly
        # Ideally, we need to figure out how to distinguish whether an attribute
        # workflow = Workflow(...)
        # workflow.foo
        # workflow.bar
        # is an explicit input and/or
        # workflow.workflowname__step__1__stepname
        # an auto-generated internal name
        # (See https://workflow-inference-compiler.readthedocs.io/en/latest/dev/algorithms.html#namespacing)
        # or if it is just a regular python field/method, etc.
        # workflow.__repr__
        # (and thus has nothing to do with CWL and should NOT be added as an input)
        # For now, let's completely ignore that distinction, and hopefully
        # the user never wants to get or set built-in python attributes!
        logger.warning(f'Adding a new input {__name} to workflow {self.process_name}')
        self._input_names.append(__name)
        inp = ProcessInput(__name, 'Any')  # Use Any type
        inp.parent_obj = self  # Create a reference to the parent Workflow object
        self.inputs.append(inp)
        return inp

    def add_output(self, __name: str) -> Any:
        logger.warning(f'Adding a new output {__name} to workflow {self.process_name}')
        self._output_names.append(__name)
        out = ProcessOutput(__name, 'Any')  # Use Any type
        out.parent_obj = self  # Create a reference to the parent Workflow object
        self.outputs.append(out)
        return out

    # def __repr__(self) -> str:
    #     repr_ = ...
    #     return repr_

    def __setattr__(self, __name: str, __value: Any) -> Any:
        if __name in ["inputs", "outputs", "yaml", "cfg_yaml", "process_name", "_input_names", "_output_names",
                      "__private_attributes__", "__pydantic_private__"]:
            return super().__setattr__(__name, __value)
        # NOTE: Unlike the syntax `... = workflow.attribute`
        # where attribute is interpreted to be a formal parameter / an input variable for the workflow,
        # the syntax `workflow.attribute = ...` is ambiguous.
        # attribute could refer to the actual parameter / input value, or
        # attribute could refer to a formal output variable for the workflow.
        # By default, we have arbitrarily chosen the former.
        if hasattr(self, "_input_names"):
            if __name not in self._input_names:
                self.add_input(__name)
            set_input_Step_Workflow(self, __name, __value)
        else:
            return super().__setattr__(__name, __value)

    def __getattr__(self, __name: str) -> Any:
        if __name in ["__pydantic_private__", "__class__", "__private_attributes__"]:
            return super().__getattribute__(__name)
        # TODO: double check the following logic.
        if __name != "_input_names" and __name != "_output_names":  # and hasattr(self, "_output_names") ?
            if __name in self._output_names:
                return self.outputs[self._output_names.index(__name)]
            else:
                return self.add_output(__name)
        # https://github.com/pydantic/pydantic/blob/812516d71a8696d5e29c5bdab40336d82ccde412/pydantic/main.py#L743-744
        # "We put `__getattr__` in a non-TYPE_CHECKING block because otherwise, mypy allows arbitrary attribute access
        # The same goes for __setattr__ and __delattr__, see: https://github.com/pydantic/pydantic/issues/8643"
        return super().__getattr__(__name)  # type: ignore

    def _save_yaml(self) -> None:
        _WIC_PATH.mkdir(parents=True, exist_ok=True)
        self.yml_path = _WIC_PATH.joinpath(f"{self.process_name}.wic")
        with open(f"{self.process_name}.wic", "w", encoding="utf-8") as file:
            file.write(yaml.dump(self.yaml))

    def _save_all_cwl(self) -> None:
        """Save CWL files to cwl_adapters.

        This is necessary for WIC to compile the workflow.
        """
        _WIC_PATH.mkdir(parents=True, exist_ok=True)
        for s in self.steps:
            try:
                if isinstance(s, Step):
                    s._save_cwl(_WIC_PATH)  # pylint: disable=W0212
            except BaseException as exc:
                raise exc

    def flatten_steps(self) -> list[Step]:
        """Flattens all Steps into a linear list. This is similar, but different, from inlineing.

        Returns:
            list[Step]: a linear list of Steps
        """
        steps = []
        for step in self.steps:
            if isinstance(step, Step):
                steps.append(step)
            if isinstance(step, Workflow):
                steps += step.flatten_steps()
        return steps

    # NOTE: Cannot return list[Workflow] because Workflow is not yet defined.
    def flatten_subworkflows(self) -> list:
        """Flattens all sub-Workflows into a linear list. The root workflow will be at index 0,
        i.e. to get a list of all proper subworkflows, just use [1:]

        Returns:
            list: a linear list of sub-Workflows.
        """
        subworkflows = [self]
        for step in self.steps:
            if isinstance(step, Workflow):
                subworkflows += step.flatten_subworkflows()
        return subworkflows

    def compile(self, write_to_disk: bool = False) -> CompilerInfo:
        """Compile Workflow using WIC.

        Args:
            write_to_disk (bool, optional): Save the compiled CWL Workflow to disk. Defaults to False.

        Returns:
            CompilerInfo: Contains the data associated with compiled subworkflows\n
            (in the Rose Tree) together with mutable cumulative environment\n
            information which needs to be passed through the recursion.
        """
        global global_config
        self._validate()
        args = get_args(self.process_name)  # Use mock CLI args

        graph = get_graph_reps(self.process_name)
        yaml_tree = YamlTree(StepId(self.process_name, 'global'), self.yaml)

        # NOTE: This is critical for control flow in the compiler. See utils.get_subkeys()
        steps_config = extract_tools_paths_NONPORTABLE(self.flatten_steps())
        global_config = merge(steps_config, global_config, strategy=Strategy.TYPESAFE_REPLACE)

        # The compile_workflow function is 100% in-memory
        compiler_info = compiler.compile_workflow(yaml_tree, args, [], [graph], {}, {}, {}, {},
                                                  global_config, True, relative_run_path=True, testing=False)

        if write_to_disk:
            # Now we can choose whether to write_to_disk or not
            rose_tree: RoseTree = compiler_info.rose
            input_output.write_to_disk(rose_tree, Path('autogenerated/'), True, args.inputs_file)

        return compiler_info

    def run(self) -> None:
        """Run compiled workflow."""
        logger.info(f"Running {self.process_name}")
        plugins.logging_filters()
        compiler_info = self.compile(write_to_disk=True)

        args = get_args(self.process_name)  # Use mock CLI args
        rose_tree: RoseTree = compiler_info.rose

        # cwl-docker-extract recursively `docker pull`s all images in all subworkflows.
        # This is important because cwltool only uses `docker run` when executing
        # workflows, and if there is a local image available,
        # `docker run` will NOT query the remote repository for the latest image!
        # cwltool has a --force-docker-pull option, but this may cause multiple pulls in parallel.
        if args.container_engine == 'singularity':
            cmd = ['cwl-docker-extract', f'-s --dir {args.homedir}', f'autogenerated/{self.process_name}.cwl']
        else:
            cmd = ['cwl-docker-extract', '--force-download', f'autogenerated/{self.process_name}.cwl']
        sub.run(cmd, check=True)

        # If you don't like it, you can programmatically overwrite anything in args
        # args.docker_remove_entrypoints = True
        if args.docker_remove_entrypoints:
            # Requires root, so guard behind CLI option
            if args.container_engine == 'docker':
                plugins.remove_entrypoints_docker()
            if args.container_engine == 'podman':
                plugins.remove_entrypoints_podman()

            rose_tree = plugins.dockerPull_append_noentrypoint_rosetree(rose_tree)
            input_output.write_to_disk(rose_tree, Path('autogenerated/'), True, args.inputs_file)

        # Do NOT capture stdout and/or stderr and pipe warnings and errors into a black hole.
        retval = run_local_module.run_local(args, rose_tree, args.cachedir, 'cwltool', True)

        # Finally, since there is an output file copying bug in cwltool,
        # we need to copy the output files manually. See comment above.
        if args.copy_output_files:
            run_local_module.copy_output_files(self.process_name)

# Process = Union[Step, Workflow]



import gc
import subprocess as sub
import time

# import matplotlib.pyplot as plt
import numpy as np
from scipy.optimize import least_squares
import yaml

from sophios.utils_yaml import wic_loader


def test_scattering_scaling(minval: int = 100, maxval: int = 500, step: int = 100) -> None:
    """Tests that execution time scales linearly with input array size"""
    # First compile the workflow
    cmd = ['sophios', '--yaml', 'examples/scattering_scaling.wic']
    sub.run(cmd, check=True)

    basedir = 'autogenerated'
    times = {}
    for i in range(minval, maxval + 1, step):
        # Modify the inputs file using the scaling parameter
        inputs_yml = f'{basedir}/scattering_scaling_inputs.yml'
        with open(inputs_yml, mode='r', encoding='utf-8') as f:
            d = yaml.load(f.read(), Loader=wic_loader())

        d['scattering_scaling__step__1__array_int___maxval'] = i

        with open(inputs_yml, mode='w', encoding='utf-8') as f:
            f.write(yaml.dump(d))

        # Garbage collect now, to avoid disturbing the workflow execution timings
        gc.collect()

        # Time the workflow execution
        t1 = time.time()
        cwl_workflow = f'{basedir}/scattering_scaling.cwl'
        cmd = ['cwltool', '--parallel', cwl_workflow, inputs_yml]
        # NOTE: Do NOT use toil for this test, simply because it is >10X slower
        # See https://github.com/DataBiosphere/toil/issues/4434
        # cmd = ['toil-cwl-runner', cwl_workflow, inputs_yml]
        # sub.run(cmd) # for debugging
        sub.run(cmd, check=True, stdout=sub.PIPE, stderr=sub.STDOUT)
        t2 = time.time()
        dt = t2 - t1

        times[i] = dt
        print('i, dt', i, dt)

    with open('times.csv', mode='w', encoding='utf-8') as t:
        for i_, dt_ in times.items():
            t.write(f'{i_}, {dt_}\n')

    # plt.scatter(times.keys(), times.values())
    # plt.savefig('times.png')

    # Fit the data to a quadratic model and check that the
    # leading coefficient is zero (within some tolerance).
    def quadratic(params: np.ndarray, x: float) -> float:
        return float(params[0]*x*x + params[1]*x + params[2])

    def residuals(params: np.ndarray, xs: np.ndarray, ys: np.ndarray) -> np.ndarray:
        ys_model = np.array([quadratic(params, x) for x in xs])
        res: np.ndarray = ys - ys_model
        return res

    xs = np.array(list(times.keys()))
    ys = np.array(list(times.values()))
    init_guess_linear = [0.0, 1.0, 0.0]  # y = x
    result = least_squares(residuals, init_guess_linear, args=(xs, ys))

    print(result)
    if not result.success:
        print('Warning! least_squares optimization did not succeed!')

    tol_quad = 0.001  # adjust as necessary
    is_linear = abs(result.x[0]) < tol_quad
    if not is_linear:
        print('Warning! fit is not linear!')
    assert is_linear


if __name__ == '__main__':
    test_scattering_scaling(100, 500, 100)



from typing import Any, Dict, List, NamedTuple, Tuple

import networkx as nx

# See https://mypy.readthedocs.io/en/stable/kinds_of_types.html#type-aliases

# See https://www.sphinx-doc.org/en/master/usage/extensions/autodoc.html#confval-autodoc_type_aliases
# See https://www.sphinx-doc.org/en/master/usage/extensions/napoleon.html#confval-napoleon_type_aliases
# The sphinx autodoc documentation claims type aliases can be added to
# autodoc_type_aliases in docs/conf.py instead of showing their expansions.
# However, I can't seem to get it to work.
# TODO: Consider removing all type aliases in favor of classes.

KV = Dict[str, Any]
Cwl = KV
Json = KV
Yaml = KV

# In python there are unfortunately an enormous number of ways to represent the humble struct.
# See https://stackoverflow.com/questions/53409117/what-are-the-main-differences-of-namedtuple-and-typeddict-in-python-mypy
# I have chosen to use NamedTuple to emphasize the immutability aspect (see below).
# See https://mypy.readthedocs.io/en/stable/kinds_of_types.html#named-tuples


class Tool(NamedTuple):
    run_path: str
    cwl: Cwl


class StepId(NamedTuple):
    stem: str  # filename without extension
    plugin_ns: str  # left column of yml_paths.txt


Tools = Dict[StepId, Tool]

# NOTE: Please read the Namespacing section of docs/devguide.md !!!
Namespace = str
Namespaces = List[Namespace]

WorkflowInputs = Dict[str, Dict[str, str]]
WorkflowInputsFile = Dict[str, Dict[str, str]]
WorkflowOutputs = List[Yaml]
InternalOutputs = List[str]
ExplicitEdgeDef = Tuple[Namespaces, str]
ExplicitEdgeDefs = Dict[str, ExplicitEdgeDef]
ExplicitEdgeCalls = Dict[str, ExplicitEdgeDef]
PluginID = int
StepName1 = str
DiGraph = Any  # graphviz.DiGraph


class GraphData():
    # pylint:disable=too-few-public-methods
    def __init__(self,
                 name: str,  # TODO: Should this be StepId?
                 nodes: List[Tuple[str, Dict]] = [],
                 edges: List[Tuple[str, str, Dict]] = [],
                 subgraphs: List[Any] = [],
                 ranksame: List[str] = []) -> None:
        # NOTE: See comments in utils_graphs.flatten_graphdata() !!!
        self.name = name
        self.nodes = nodes
        self.edges = edges
        self.subgraphs = subgraphs
        self.ranksame = ranksame


# This groups together the classes which represent our graph.
# Excluding --graph_inline_depth related code, all graph
# operations should be performed on all representations.
class GraphReps(NamedTuple):
    graphviz: DiGraph
    networkx: nx.DiGraph
    graphdata: GraphData


YamlDSLArgs = Yaml

# Since we cannot store extra tags in CWL files, we need a data structure
# to store temporary compiler info that gets passed through the recursion.
# The number of subworkflows is arbitrary (zero or more), so what we want is a
# Rose Tree https://en.wikipedia.org/wiki/Rose_tree
# Unfortunately, since mypy does not support Algebraic Data Types (ADTs)
# we have to break the recursion by replacing the recursive instance of RoseTree with Any :(
DataType = Any


class RoseTree(NamedTuple):
    data: DataType
    sub_trees: List[Any]  # Any = RoseTree
# Note that instead of DataType we could provide a specific type, but remember that
# a Rose Tree is defined by its structure, not by the specific type of data it contains.
# We can simply cast to a specific type at each call site, i.e.
# data: SpecificType = rose_tree.data

# Now we can define a specific data type for a single node in our Abstract Syntax Tree.


class NodeData(NamedTuple):
    namespaces: Namespaces
    name: str
    yml: Yaml  # i.e. The AST that was compiled.
    # If this is not the AST that was passed in, then the compiler introduced
    # some modifications (i.e. --insert_steps_automatically) and you need to recompile
    compiled_cwl: Cwl
    tool: Tool
    workflow_inputs_file: WorkflowInputsFile
    explicit_edge_defs: ExplicitEdgeDefs
    explicit_edge_calls: ExplicitEdgeCalls
    graph: GraphReps
    inputs_workflow: WorkflowInputs
    step_name_1: StepName1


class EnvData(NamedTuple):
    input_mapping: Dict[str, List[str]]
    output_mapping: Dict[str, str]
    inputs_file_workflow: WorkflowInputsFile
    vars_workflow_output_internal: InternalOutputs
    explicit_edge_defs: ExplicitEdgeDefs
    explicit_edge_calls: ExplicitEdgeCalls


class CompilerInfo(NamedTuple):
    rose: RoseTree
    env: EnvData
# Note that while Tuples and NamedTuples are immutable, they can contain mutable entries!
# Let's partition the data into entries which are immutable / fixed
# (well, at least after compilation of the subworkflow is complete) and entries
# which are mutably updated throughout the recursion. The latter are essentially
# environment variables of sorts, and they do not need to be stored in the Rose Tree.

# Create a type for our Abstract Syntax Tree (AST).
# We can probably use Dict here if str is step_name_i not just yaml_stem.
# If we need to insert steps, that will happen
# after edge inference, so there should not be a uniqueness issue w.r.t. step
# number re-indexing.


class YamlTree(NamedTuple):
    step_id: StepId
    yml: Yaml


class YamlForest(NamedTuple):
    yaml_tree: YamlTree
    sub_forests: List[Tuple[StepId, Any]]  # Any = YamlForest
